# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from functools import wraps
import types
import warnings

import cf_units
import numpy as np

import iris
import iris.coords as coords
import iris.tests.stock

# Run tests in no graphics mode if matplotlib is not available.
if tests.MPL_AVAILABLE:
    import matplotlib.pyplot as plt
    import iris.plot as iplt
    import iris.quickplot as qplt
    import iris.symbols


@tests.skip_data
def simple_cube():
    cube = iris.tests.stock.realistic_4d()
    cube = cube[:, 0, 0, :]
    cube.coord("time").guess_bounds()
    return cube


@tests.skip_plot
class TestSimple(tests.GraphicsTest):
    def test_points(self):
        cube = simple_cube()
        qplt.contourf(cube)
        self.check_graphic()

    def test_bounds(self):
        cube = simple_cube()
        qplt.pcolor(cube)
        self.check_graphic()


@tests.skip_plot
class TestMissingCoord(tests.GraphicsTest):
    def _check(self, cube):
        qplt.contourf(cube)
        self.check_graphic()

        qplt.pcolor(cube)
        self.check_graphic()

    def test_no_u(self):
        cube = simple_cube()
        cube.remove_coord("grid_longitude")
        self._check(cube)

    def test_no_v(self):
        cube = simple_cube()
        cube.remove_coord("time")
        self._check(cube)

    def test_none(self):
        cube = simple_cube()
        cube.remove_coord("grid_longitude")
        cube.remove_coord("time")
        self._check(cube)


@tests.skip_data
@tests.skip_plot
class TestMissingCS(tests.GraphicsTest):
    @tests.skip_data
    def test_missing_cs(self):
        cube = tests.stock.simple_pp()
        cube.coord("latitude").coord_system = None
        cube.coord("longitude").coord_system = None
        qplt.contourf(cube)
        qplt.plt.gca().coastlines("110m")
        self.check_graphic()


@tests.skip_plot
@tests.skip_data
class TestHybridHeight(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        self.cube = iris.tests.stock.realistic_4d()[0, :15, 0, :]

    def _check(self, plt_method, test_altitude=True):
        plt_method(self.cube)
        self.check_graphic()

        plt_method(self.cube, coords=["level_height", "grid_longitude"])
        self.check_graphic()

        plt_method(self.cube, coords=["grid_longitude", "level_height"])
        self.check_graphic()

        if test_altitude:
            plt_method(self.cube, coords=["grid_longitude", "altitude"])
            self.check_graphic()

            plt_method(self.cube, coords=["altitude", "grid_longitude"])
            self.check_graphic()

    def test_points(self):
        self._check(qplt.contourf)

    def test_bounds(self):
        self._check(qplt.pcolor, test_altitude=False)

    def test_orography(self):
        qplt.contourf(self.cube)
        iplt.orography_at_points(self.cube)
        iplt.points(self.cube)
        self.check_graphic()

        coords = ["altitude", "grid_longitude"]
        qplt.contourf(self.cube, coords=coords)
        iplt.orography_at_points(self.cube, coords=coords)
        iplt.points(self.cube, coords=coords)
        self.check_graphic()

        # TODO: Test bounds once they are supported.
        with self.assertRaises(NotImplementedError):
            qplt.pcolor(self.cube)
            iplt.orography_at_bounds(self.cube)
            iplt.outline(self.cube)
            self.check_graphic()


@tests.skip_plot
@tests.skip_data
class Test1dPlotMultiArgs(tests.GraphicsTest):
    # tests for iris.plot using multi-argument calling convention

    def setUp(self):
        super().setUp()
        self.cube1d = _load_4d_testcube()[0, :, 0, 0]
        self.draw_method = iplt.plot

    def test_cube(self):
        # just plot a cube against its dim coord
        self.draw_method(self.cube1d)  # altitude vs temp
        self.check_graphic()

    def test_coord(self):
        # plot the altitude coordinate
        self.draw_method(self.cube1d.coord("altitude"))
        self.check_graphic()

    def test_coord_cube(self):
        # plot temperature against sigma
        self.draw_method(self.cube1d.coord("sigma"), self.cube1d)
        self.check_graphic()

    def test_cube_coord(self):
        # plot a vertical profile of temperature
        self.draw_method(self.cube1d, self.cube1d.coord("altitude"))
        self.check_graphic()

    def test_coord_coord(self):
        # plot two coordinates that are not mappable
        self.draw_method(
            self.cube1d.coord("sigma"), self.cube1d.coord("altitude")
        )
        self.check_graphic()

    def test_coord_coord_map(self):
        # plot lat-lon aux coordinates of a trajectory, which draws a map
        lon = iris.coords.AuxCoord(
            [0, 5, 10, 15, 20, 25, 30, 35, 40, 45],
            standard_name="longitude",
            units="degrees_north",
        )
        lat = iris.coords.AuxCoord(
            [45, 55, 50, 60, 55, 65, 60, 70, 65, 75],
            standard_name="latitude",
            units="degrees_north",
        )
        self.draw_method(lon, lat)
        plt.gca().coastlines("110m")
        self.check_graphic()

    def test_cube_cube(self):
        # plot two phenomena against eachother, in this case just dummy data
        cube1 = self.cube1d.copy()
        cube2 = self.cube1d.copy()
        cube1.rename("some phenomenon")
        cube2.rename("some other phenomenon")
        cube1.units = cf_units.Unit("no_unit")
        cube2.units = cf_units.Unit("no_unit")
        cube1.data[:] = np.linspace(0, 1, 7)
        cube2.data[:] = np.exp(cube1.data)
        self.draw_method(cube1, cube2)
        self.check_graphic()

    def test_incompatible_objects(self):
        # incompatible objects (not the same length) should raise an error
        with self.assertRaises(ValueError):
            self.draw_method(self.cube1d.coord("time"), (self.cube1d))

    def test_multimidmensional(self):
        # multidimensional cubes are not allowed
        cube = _load_4d_testcube()[0, :, :, 0]
        with self.assertRaises(ValueError):
            self.draw_method(cube)

    def test_not_cube_or_coord(self):
        # inputs must be cubes or coordinates, otherwise an error should be
        # raised
        xdim = np.arange(self.cube1d.shape[0])
        with self.assertRaises(TypeError):
            self.draw_method(xdim, self.cube1d)

    def test_plot_old_coords_kwarg(self):
        # Coords used to be a valid kwarg to plot, but it was deprecated and
        # we are maintaining a reasonable exception, check that it is raised
        # here.
        with self.assertRaises(TypeError):
            self.draw_method(self.cube1d, coords=None)


@tests.skip_plot
class Test1dQuickplotPlotMultiArgs(Test1dPlotMultiArgs):
    # tests for iris.plot using multi-argument calling convention

    def setUp(self):
        tests.GraphicsTest.setUp(self)
        self.cube1d = _load_4d_testcube()[0, :, 0, 0]
        self.draw_method = qplt.plot


@tests.skip_data
@tests.skip_plot
class Test1dScatter(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        self.cube = iris.load_cube(
            tests.get_data_path(("NAME", "NAMEIII_trajectory.txt")),
            "Temperature",
        )
        self.draw_method = iplt.scatter

    def test_coord_coord(self):
        x = self.cube.coord("longitude")
        y = self.cube.coord("altitude")
        c = self.cube.data
        self.draw_method(x, y, c=c, edgecolor="none")
        self.check_graphic()

    def test_coord_coord_map(self):
        x = self.cube.coord("longitude")
        y = self.cube.coord("latitude")
        c = self.cube.data
        self.draw_method(x, y, c=c, edgecolor="none")
        plt.gca().coastlines("110m")
        self.check_graphic()

    def test_coord_cube(self):
        x = self.cube.coord("latitude")
        y = self.cube
        c = self.cube.coord("Travel Time").points
        self.draw_method(x, y, c=c, edgecolor="none")
        self.check_graphic()

    def test_cube_coord(self):
        x = self.cube
        y = self.cube.coord("altitude")
        c = self.cube.coord("Travel Time").points
        self.draw_method(x, y, c=c, edgecolor="none")
        self.check_graphic()

    def test_cube_cube(self):
        x = iris.load_cube(
            tests.get_data_path(("NAME", "NAMEIII_trajectory.txt")),
            "Rel Humidity",
        )
        y = self.cube
        c = self.cube.coord("Travel Time").points
        self.draw_method(x, y, c=c, edgecolor="none")
        self.check_graphic()

    def test_incompatible_objects(self):
        # cubes/coordinates of different sizes cannot be plotted
        x = self.cube
        y = self.cube.coord("altitude")[:-1]
        with self.assertRaises(ValueError):
            self.draw_method(x, y)

    def test_multidimensional(self):
        # multidimensional cubes/coordinates are not allowed
        x = _load_4d_testcube()[0, :, :, 0]
        y = x.coord("model_level_number")
        with self.assertRaises(ValueError):
            self.draw_method(x, y)

    def test_not_cube_or_coord(self):
        # inputs must be cubes or coordinates
        x = np.arange(self.cube.shape[0])
        y = self.cube
        with self.assertRaises(TypeError):
            self.draw_method(x, y)


@tests.skip_data
@tests.skip_plot
class Test1dQuickplotScatter(Test1dScatter):
    def setUp(self):
        tests.GraphicsTest.setUp(self)
        self.cube = iris.load_cube(
            tests.get_data_path(("NAME", "NAMEIII_trajectory.txt")),
            "Temperature",
        )
        self.draw_method = qplt.scatter


@tests.skip_data
@tests.skip_plot
class TestAttributePositive(tests.GraphicsTest):
    def test_1d_positive_up(self):
        path = tests.get_data_path(("NetCDF", "ORCA2", "votemper.nc"))
        cube = iris.load_cube(path)
        qplt.plot(cube.coord("depth"), cube[0, :, 60, 80])
        self.check_graphic()

    def test_1d_positive_down(self):
        path = tests.get_data_path(("NetCDF", "ORCA2", "votemper.nc"))
        cube = iris.load_cube(path)
        qplt.plot(cube[0, :, 60, 80], cube.coord("depth"))
        self.check_graphic()

    def test_2d_positive_up(self):
        path = tests.get_data_path(
            ("NetCDF", "testing", "small_theta_colpex.nc")
        )
        cube = iris.load_cube(path, "air_potential_temperature")[0, :, 42, :]
        qplt.pcolormesh(cube)
        self.check_graphic()

    def test_2d_positive_down(self):
        path = tests.get_data_path(("NetCDF", "ORCA2", "votemper.nc"))
        cube = iris.load_cube(path)[0, :, 42, :]
        qplt.pcolormesh(cube)
        self.check_graphic()


# Caches _load_4d_testcube so subsequent calls are faster
def cache(fn, cache={}):
    def inner(*args, **kwargs):
        key = fn.__name__
        if key not in cache:
            cache[key] = fn(*args, **kwargs)
        return cache[key]

    return inner


@cache
@tests.skip_data
def _load_4d_testcube():
    # Load example 4d data (TZYX).
    test_cube = iris.tests.stock.realistic_4d()
    # Replace forecast_period coord with a multi-valued version.
    time_coord = test_cube.coord("time")
    n_times = len(time_coord.points)
    forecast_dims = test_cube.coord_dims(time_coord)
    test_cube.remove_coord("forecast_period")
    # Make up values (including bounds), to roughly match older testdata.
    point_values = np.linspace((1 + 1.0 / 6), 2.0, n_times)
    point_uppers = point_values + (point_values[1] - point_values[0])
    bound_values = np.column_stack([point_values, point_uppers])
    # NOTE: this must be a DimCoord
    #  - an equivalent AuxCoord produces different plots.
    new_forecast_coord = iris.coords.DimCoord(
        points=point_values,
        bounds=bound_values,
        standard_name="forecast_period",
        units=cf_units.Unit("hours"),
    )
    test_cube.add_aux_coord(new_forecast_coord, forecast_dims)
    # Heavily reduce dimensions for faster testing.
    # NOTE: this makes ZYX non-contiguous.  Doesn't seem to matter for now.
    test_cube = test_cube[:, ::10, ::10, ::10]
    return test_cube


@cache
def _load_wind_no_bounds():
    # Load the COLPEX data => TZYX
    path = tests.get_data_path(("PP", "COLPEX", "small_eastward_wind.pp"))
    wind = iris.load_cube(path, "x_wind")

    # Remove bounds from all coords that have them.
    wind.coord("grid_latitude").bounds = None
    wind.coord("grid_longitude").bounds = None
    wind.coord("level_height").bounds = None
    wind.coord("sigma").bounds = None

    return wind[:, :, :50, :50]


def _time_series(src_cube):
    # Until we have plotting support for multiple axes on the same dimension,
    # remove the time coordinate and its axis.
    cube = src_cube.copy()
    cube.remove_coord("time")
    return cube


def _date_series(src_cube):
    # Until we have plotting support for multiple axes on the same dimension,
    # remove the forecast_period coordinate and its axis.
    cube = src_cube.copy()
    cube.remove_coord("forecast_period")
    return cube


@tests.skip_plot
class SliceMixin:
    """Mixin class providing tests for each 2-dimensional permutation of axes.

    Requires self.draw_method to be the relevant plotting function,
    and self.results to be a dictionary containing the desired test results."""

    def test_yx(self):
        cube = self.wind[0, 0, :, :]
        self.draw_method(cube)
        self.check_graphic()

    def test_zx(self):
        cube = self.wind[0, :, 0, :]
        self.draw_method(cube)
        self.check_graphic()

    def test_tx(self):
        cube = _time_series(self.wind[:, 0, 0, :])
        self.draw_method(cube)
        self.check_graphic()

    def test_zy(self):
        cube = self.wind[0, :, :, 0]
        self.draw_method(cube)
        self.check_graphic()

    def test_ty(self):
        cube = _time_series(self.wind[:, 0, :, 0])
        self.draw_method(cube)
        self.check_graphic()

    def test_tz(self):
        cube = _time_series(self.wind[:, :, 0, 0])
        self.draw_method(cube)
        self.check_graphic()


@tests.skip_data
class TestContour(tests.GraphicsTest, SliceMixin):
    """Test the iris.plot.contour routine."""

    def setUp(self):
        super().setUp()
        self.wind = _load_4d_testcube()
        self.draw_method = iplt.contour


@tests.skip_data
class TestContourf(tests.GraphicsTest, SliceMixin):
    """Test the iris.plot.contourf routine."""

    def setUp(self):
        super().setUp()
        self.wind = _load_4d_testcube()
        self.draw_method = iplt.contourf


@tests.skip_data
class TestPcolor(tests.GraphicsTest, SliceMixin):
    """Test the iris.plot.pcolor routine."""

    def setUp(self):
        super().setUp()
        self.wind = _load_4d_testcube()
        self.draw_method = iplt.pcolor


@tests.skip_data
class TestPcolormesh(tests.GraphicsTest, SliceMixin):
    """Test the iris.plot.pcolormesh routine."""

    def setUp(self):
        super().setUp()
        self.wind = _load_4d_testcube()
        self.draw_method = iplt.pcolormesh


def check_warnings(method):
    """
    Decorator that adds a catch_warnings and filter to assert
    the method being decorated issues a UserWarning.

    """

    @wraps(method)
    def decorated_method(self, *args, **kwargs):
        # Force reset of iris.coords warnings registry to avoid suppression of
        # repeated warnings. warnings.resetwarnings() does not do this.
        if hasattr(coords, "__warningregistry__"):
            coords.__warningregistry__.clear()

        # Check that method raises warning.
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            with self.assertRaises(UserWarning):
                return method(self, *args, **kwargs)

    return decorated_method


def ignore_warnings(method):
    """
    Decorator that adds a catch_warnings and filter to suppress
    any warnings issues by the method being decorated.

    """

    @wraps(method)
    def decorated_method(self, *args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return method(self, *args, **kwargs)

    return decorated_method


class CheckForWarningsMetaclass(type):
    """
    Metaclass that adds a further test for each base class test
    that checks that each test raises a UserWarning. Each base
    class test is then overriden to ignore warnings in order to
    check the underlying functionality.

    """

    def __new__(cls, name, bases, local):
        def add_decorated_methods(attr_dict, target_dict, decorator):
            for key, value in attr_dict.items():
                if isinstance(value, types.FunctionType) and key.startswith(
                    "test"
                ):
                    new_key = "_".join((key, decorator.__name__))
                    if new_key not in target_dict:
                        wrapped = decorator(value)
                        wrapped.__name__ = new_key
                        target_dict[new_key] = wrapped
                    else:
                        raise RuntimeError(
                            "A attribute called {!r} "
                            "already exists.".format(new_key)
                        )

        def override_with_decorated_methods(attr_dict, target_dict, decorator):
            for key, value in attr_dict.items():
                if isinstance(value, types.FunctionType) and key.startswith(
                    "test"
                ):
                    target_dict[key] = decorator(value)

        # Add decorated versions of base methods
        # to check for warnings.
        for base in bases:
            add_decorated_methods(base.__dict__, local, check_warnings)

        # Override base methods to ignore warnings.
        for base in bases:
            override_with_decorated_methods(
                base.__dict__, local, ignore_warnings
            )

        return type.__new__(cls, name, bases, local)


@tests.skip_data
@tests.iristest_timing_decorator
class TestPcolorNoBounds(
    tests.GraphicsTest_nometa, SliceMixin, metaclass=CheckForWarningsMetaclass
):
    """
    Test the iris.plot.pcolor routine on a cube with coordinates
    that have no bounds.

    """

    def setUp(self):
        super().setUp()
        self.wind = _load_wind_no_bounds()
        self.draw_method = iplt.pcolor


@tests.skip_data
@tests.iristest_timing_decorator
class TestPcolormeshNoBounds(
    tests.GraphicsTest_nometa, SliceMixin, metaclass=CheckForWarningsMetaclass
):
    """
    Test the iris.plot.pcolormesh routine on a cube with coordinates
    that have no bounds.

    """

    def setUp(self):
        super().setUp()
        self.wind = _load_wind_no_bounds()
        self.draw_method = iplt.pcolormesh


@tests.skip_plot
class Slice1dMixin:
    """Mixin class providing tests for each 1-dimensional permutation of axes.

    Requires self.draw_method to be the relevant plotting function,
    and self.results to be a dictionary containing the desired test results."""

    def test_x(self):
        cube = self.wind[0, 0, 0, :]
        self.draw_method(cube)
        self.check_graphic()

    def test_y(self):
        cube = self.wind[0, 0, :, 0]
        self.draw_method(cube)
        self.check_graphic()

    def test_z(self):
        cube = self.wind[0, :, 0, 0]
        self.draw_method(cube)
        self.check_graphic()

    def test_t(self):
        cube = _time_series(self.wind[:, 0, 0, 0])
        self.draw_method(cube)
        self.check_graphic()

    def test_t_dates(self):
        cube = _date_series(self.wind[:, 0, 0, 0])
        self.draw_method(cube)
        plt.gcf().autofmt_xdate()
        plt.xlabel("Phenomenon time")

        self.check_graphic()


@tests.skip_data
class TestPlot(tests.GraphicsTest, Slice1dMixin):
    """Test the iris.plot.plot routine."""

    def setUp(self):
        super().setUp()
        self.wind = _load_4d_testcube()
        self.draw_method = iplt.plot


@tests.skip_data
class TestQuickplotPlot(tests.GraphicsTest, Slice1dMixin):
    """Test the iris.quickplot.plot routine."""

    def setUp(self):
        super().setUp()
        self.wind = _load_4d_testcube()
        self.draw_method = qplt.plot


_load_cube_once_cache = {}


def load_cube_once(filename, constraint):
    """Same syntax as load_cube, but will only load a file once,

    then cache the answer in a dictionary.

    """
    global _load_cube_once_cache
    key = (filename, str(constraint))
    cube = _load_cube_once_cache.get(key, None)

    if cube is None:
        cube = iris.load_cube(filename, constraint)
        _load_cube_once_cache[key] = cube

    return cube


class LambdaStr:
    """Provides a callable function which has a sensible __repr__."""

    def __init__(self, repr, lambda_fn):
        self.repr = repr
        self.lambda_fn = lambda_fn

    def __call__(self, *args, **kwargs):
        return self.lambda_fn(*args, **kwargs)

    def __repr__(self):
        return self.repr


@tests.skip_data
@tests.skip_plot
class TestPlotCoordinatesGiven(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        filename = tests.get_data_path(
            ("PP", "COLPEX", "theta_and_orog_subset.pp")
        )
        self.cube = load_cube_once(filename, "air_potential_temperature")

        self.draw_module = iris.plot
        self.contourf = LambdaStr(
            "iris.plot.contourf",
            lambda cube, *args, **kwargs: iris.plot.contourf(
                cube, *args, **kwargs
            ),
        )
        self.contour = LambdaStr(
            "iris.plot.contour",
            lambda cube, *args, **kwargs: iris.plot.contour(
                cube, *args, **kwargs
            ),
        )
        self.points = LambdaStr(
            "iris.plot.points",
            lambda cube, *args, **kwargs: iris.plot.points(
                cube, c=cube.data, *args, **kwargs
            ),
        )
        self.plot = LambdaStr(
            "iris.plot.plot",
            lambda cube, *args, **kwargs: iris.plot.plot(
                cube, *args, **kwargs
            ),
        )

        self.results = {
            "yx": (
                [self.contourf, ["grid_latitude", "grid_longitude"]],
                [self.contourf, ["grid_longitude", "grid_latitude"]],
                [self.contour, ["grid_latitude", "grid_longitude"]],
                [self.contour, ["grid_longitude", "grid_latitude"]],
                [self.points, ["grid_latitude", "grid_longitude"]],
                [self.points, ["grid_longitude", "grid_latitude"]],
            ),
            "zx": (
                [self.contourf, ["model_level_number", "grid_longitude"]],
                [self.contourf, ["grid_longitude", "model_level_number"]],
                [self.contour, ["model_level_number", "grid_longitude"]],
                [self.contour, ["grid_longitude", "model_level_number"]],
                [self.points, ["model_level_number", "grid_longitude"]],
                [self.points, ["grid_longitude", "model_level_number"]],
            ),
            "tx": (
                [self.contourf, ["time", "grid_longitude"]],
                [self.contourf, ["grid_longitude", "time"]],
                [self.contour, ["time", "grid_longitude"]],
                [self.contour, ["grid_longitude", "time"]],
                [self.points, ["time", "grid_longitude"]],
                [self.points, ["grid_longitude", "time"]],
            ),
            "x": ([self.plot, ["grid_longitude"]],),
            "y": ([self.plot, ["grid_latitude"]],),
        }

    def draw(self, draw_method, *args, **kwargs):
        draw_fn = getattr(self.draw_module, draw_method)
        draw_fn(*args, **kwargs)
        self.check_graphic()

    def run_tests(self, cube, results):
        for draw_method, rcoords in results:
            draw_method(cube, coords=rcoords)
            try:
                self.check_graphic()
            except AssertionError as err:
                self.fail(
                    "Draw method %r failed with coords: %r. "
                    "Assertion message: %s" % (draw_method, rcoords, err)
                )

    def run_tests_1d(self, cube, results):
        # there is a different calling convention for 1d plots
        for draw_method, rcoords in results:
            draw_method(cube.coord(rcoords[0]), cube)
            try:
                self.check_graphic()
            except AssertionError as err:
                msg = (
                    "Draw method {!r} failed with coords: {!r}. "
                    "Assertion message: {!s}"
                )
                self.fail(msg.format(draw_method, rcoords, err))

    def test_yx(self):
        test_cube = self.cube[0, 0, :, :]
        self.run_tests(test_cube, self.results["yx"])

    def test_zx(self):
        test_cube = self.cube[0, :15, 0, :]
        self.run_tests(test_cube, self.results["zx"])

    def test_tx(self):
        test_cube = self.cube[:, 0, 0, :]
        self.run_tests(test_cube, self.results["tx"])

    def test_x(self):
        test_cube = self.cube[0, 0, 0, :]
        self.run_tests_1d(test_cube, self.results["x"])

    def test_y(self):
        test_cube = self.cube[0, 0, :, 0]
        self.run_tests_1d(test_cube, self.results["y"])

    def test_badcoords(self):
        cube = self.cube[0, 0, :, :]
        draw_fn = getattr(self.draw_module, "contourf")
        self.assertRaises(
            ValueError,
            draw_fn,
            cube,
            coords=["grid_longitude", "grid_longitude"],
        )
        self.assertRaises(
            ValueError,
            draw_fn,
            cube,
            coords=["grid_longitude", "grid_longitude", "grid_latitude"],
        )
        self.assertRaises(
            iris.exceptions.CoordinateNotFoundError,
            draw_fn,
            cube,
            coords=["grid_longitude", "wibble"],
        )
        self.assertRaises(ValueError, draw_fn, cube, coords=[])
        self.assertRaises(
            ValueError,
            draw_fn,
            cube,
            coords=[
                cube.coord("grid_longitude"),
                cube.coord("grid_longitude"),
            ],
        )
        self.assertRaises(
            ValueError,
            draw_fn,
            cube,
            coords=[
                cube.coord("grid_longitude"),
                cube.coord("grid_longitude"),
                cube.coord("grid_longitude"),
            ],
        )

    def test_non_cube_coordinate(self):
        cube = self.cube[0, :, :, 0]
        pts = -100 + np.arange(cube.shape[1]) * 13
        x = coords.DimCoord(
            pts,
            standard_name="model_level_number",
            attributes={"positive": "up"},
            units="1",
        )
        self.draw("contourf", cube, coords=["grid_latitude", x])


@tests.skip_data
@tests.skip_plot
class TestPlotDimAndAuxCoordsKwarg(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        filename = tests.get_data_path(
            ("NetCDF", "rotated", "xy", "rotPole_landAreaFraction.nc")
        )
        self.cube = iris.load_cube(filename)

    def test_default(self):
        iplt.contourf(self.cube)
        plt.gca().coastlines("110m")
        self.check_graphic()

    def test_coords(self):
        # Pass in dimension coords.
        rlat = self.cube.coord("grid_latitude")
        rlon = self.cube.coord("grid_longitude")
        iplt.contourf(self.cube, coords=[rlon, rlat])
        plt.gca().coastlines("110m")
        self.check_graphic()
        # Pass in auxiliary coords.
        lat = self.cube.coord("latitude")
        lon = self.cube.coord("longitude")
        iplt.contourf(self.cube, coords=[lon, lat])
        plt.gca().coastlines("110m")
        self.check_graphic()

    def test_coord_names(self):
        # Pass in names of dimension coords.
        iplt.contourf(self.cube, coords=["grid_longitude", "grid_latitude"])
        plt.gca().coastlines("110m")
        self.check_graphic()
        # Pass in names of auxiliary coords.
        iplt.contourf(self.cube, coords=["longitude", "latitude"])
        plt.gca().coastlines("110m")
        self.check_graphic()

    def test_yx_order(self):
        # Do not attempt to draw coastlines as it is not a map.
        iplt.contourf(self.cube, coords=["grid_latitude", "grid_longitude"])
        self.check_graphic()
        iplt.contourf(self.cube, coords=["latitude", "longitude"])
        self.check_graphic()


@tests.skip_plot
class TestSymbols(tests.GraphicsTest):
    def test_cloud_cover(self):
        iplt.symbols(
            list(range(10)),
            [0] * 10,
            [iris.symbols.CLOUD_COVER[i] for i in range(10)],
            0.375,
        )
        iplt.plt.axis("off")
        self.check_graphic()


@tests.skip_plot
class TestPlottingExceptions(tests.IrisTest):
    def setUp(self):
        self.bounded_cube = tests.stock.lat_lon_cube()
        self.bounded_cube.coord("latitude").guess_bounds()
        self.bounded_cube.coord("longitude").guess_bounds()

    def test_boundmode_multidim(self):
        # Test exception translation.
        # We can't get contiguous bounded grids from multi-d coords.
        cube = self.bounded_cube
        cube.remove_coord("latitude")
        cube.add_aux_coord(
            coords.AuxCoord(
                points=cube.data, standard_name="latitude", units="degrees"
            ),
            [0, 1],
        )
        with self.assertRaises(ValueError):
            iplt.pcolormesh(cube, coords=["longitude", "latitude"])

    def test_boundmode_4bounds(self):
        # Test exception translation.
        # We can only get contiguous bounded grids with 2 bounds per point.
        cube = self.bounded_cube
        lat = coords.AuxCoord.from_coord(cube.coord("latitude"))
        lat.bounds = np.array(
            [lat.points, lat.points + 1, lat.points + 2, lat.points + 3]
        ).transpose()
        cube.remove_coord("latitude")
        cube.add_aux_coord(lat, 0)
        with self.assertRaises(ValueError):
            iplt.pcolormesh(cube, coords=["longitude", "latitude"])

    def test_different_coord_systems(self):
        cube = self.bounded_cube
        lat = cube.coord("latitude")
        lon = cube.coord("longitude")
        lat.coord_system = iris.coord_systems.GeogCS(7000000)
        lon.coord_system = iris.coord_systems.GeogCS(7000001)
        with self.assertRaises(ValueError):
            iplt.pcolormesh(cube, coords=["longitude", "latitude"])


@tests.skip_data
@tests.skip_plot
class TestPlotOtherCoordSystems(tests.GraphicsTest):
    def test_plot_tmerc(self):
        filename = tests.get_data_path(
            ("NetCDF", "transverse_mercator", "tmean_1910_1910.nc")
        )
        self.cube = iris.load_cube(filename)
        iplt.pcolormesh(self.cube[0])
        plt.gca().coastlines("110m")
        self.check_graphic()


@tests.skip_plot
class TestPlotCitation(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        self.figure = plt.figure()
        self.axes = self.figure.gca()
        self.text = (
            "Lorem ipsum dolor sit amet, consectetur adipiscing "
            "elit, sed do eiusmod tempor incididunt ut labore et "
            "dolore magna aliqua."
        )

    def test(self):
        iplt.citation(self.text)
        self.check_graphic()

    def test_figure(self):
        iplt.citation(self.text, figure=self.figure)
        self.check_graphic()

    def test_axes(self):
        iplt.citation(self.text, axes=self.axes)
        self.check_graphic()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides Creation and saving of DOT graphs for a :class:`iris.cube.Cube`.

"""

import os
import subprocess

import iris
import iris.util


_GRAPH_INDENT = " " * 4
_SUBGRAPH_INDENT = " " * 8


_DOT_CHECKED = False
_DOT_EXECUTABLE_PATH = None


def _dot_path():
    global _DOT_CHECKED, _DOT_EXECUTABLE_PATH

    if _DOT_CHECKED:
        path = _DOT_EXECUTABLE_PATH
    else:
        path = iris.config.get_option("System", "dot_path", default="dot")
        if not os.path.exists(path):
            if not os.path.isabs(path):
                try:
                    # Check PATH
                    subprocess.check_output(
                        [path, "-V"], stderr=subprocess.STDOUT
                    )
                except (OSError, subprocess.CalledProcessError):
                    path = None
            else:
                path = None
        _DOT_EXECUTABLE_PATH = path
        _DOT_CHECKED = True
    return path


#: Whether the 'dot' program is present (required for "dotpng" output).
DOT_AVAILABLE = _dot_path() is not None


def save(cube, target):
    """Save a dot representation of the cube.

    Args:

        * cube   - A :class:`iris.cube.Cube`.
        * target - A filename or open file handle.

    See also :func:`iris.io.save`.

    """
    if isinstance(target, str):
        dot_file = open(target, "wt")
    elif hasattr(target, "write"):
        if hasattr(target, "mode") and "b" in target.mode:
            raise ValueError("Target is binary")
        dot_file = target
    else:
        raise ValueError("Can only save dot to filename or filehandle")

    try:
        dot_file.write(cube_text(cube))
    finally:
        if isinstance(target, str):
            dot_file.close()


def save_png(source, target, launch=False):
    """
    Produces a "dot" instance diagram by calling dot and optionally launching the resulting image.

    Args:

        * source - A :class:`iris.cube.Cube`, or dot filename.
        * target - A filename or open file handle.
                   If passing a file handle, take care to open it for binary output.

    Kwargs:

        * launch - Display the image. Default is False.

    See also :func:`iris.io.save`.

    """
    # From cube or dot file?
    if isinstance(source, iris.cube.Cube):
        # Create dot file
        dot_file_path = iris.util.create_temp_filename(".dot")
        save(source, dot_file_path)
    elif isinstance(source, str):
        dot_file_path = source
    else:
        raise ValueError("Can only write dot png for a Cube or DOT file")

    # Create png data
    if not _dot_path():
        raise ValueError(
            'Executable "dot" not found: '
            "Review dot_path setting in site.cfg."
        )
    # To filename or open file handle?
    if isinstance(target, str):
        subprocess.call(
            [_dot_path(), "-T", "png", "-o", target, dot_file_path]
        )
    elif hasattr(target, "write"):
        if hasattr(target, "mode") and "b" not in target.mode:
            raise ValueError("Target not binary")
        subprocess.call(
            [_dot_path(), "-T", "png", dot_file_path], stdout=target
        )
    else:
        raise ValueError("Can only write dot png for a filename or writable")

    # Display?
    if launch:
        if os.name == "mac":
            subprocess.call(("open", target))
        elif os.name == "nt":
            subprocess.call(("start", target))
        elif os.name == "posix":
            subprocess.call(("firefox", target))
        else:
            raise iris.exceptions.NotYetImplementedError(
                "Unhandled operating system. The image has been created in %s"
                % target
            )

    # Remove the dot file if we created it
    if isinstance(source, iris.cube.Cube):
        os.remove(dot_file_path)


def cube_text(cube):
    """Return a DOT text representation a `iris.cube.Cube`.

    Args:

     * cube  -  The cube for which to create DOT text.

    """
    # We use r'' type string constructor as when we type \n in a string without the r'' constructor
    # we get back a new line character - this is not what DOT expects.
    # Therefore, newline characters should be created explicitly by having multi-lined strings.
    relationships = r""
    relationships_association = r""

    dimension_nodes = r"""
    subgraph clusterCubeDimensions {
        label="Cube data"
    """

    # TODO: Separate dim_coords from aux_coords.
    coord_nodes = r"""
    subgraph clusterCoords {
        label = "Coords"
"""

    coord_system_nodes = r"""
    subgraph clusterCoordSystems {
        label = "CoordSystems"
"""

    for i, size in enumerate(cube.shape):
        dimension_nodes += "\n" + _dot_node(
            _SUBGRAPH_INDENT,
            "CubeDimension_" + str(i),
            str(i),
            [("len", size)],
        )

    # Coords and their coord_systems
    coords = sorted(cube.coords(), key=lambda c: c.name())
    written_cs = []
    for i, coord in enumerate(coords):
        coord_label = "Coord_" + str(i)
        coord_nodes += _coord_text(coord_label, coord)
        cs = coord.coord_system
        if cs:
            # Create the cs node - or find an identical, already written cs
            if cs not in written_cs:
                written_cs.append(cs)
                uid = written_cs.index(cs)
                coord_system_nodes += _coord_system_text(cs, uid)
            else:
                uid = written_cs.index(cs)
            relationships += '\n    "%s" -> "CoordSystem_%s_%s"' % (
                coord_label,
                coord.coord_system.__class__.__name__,
                uid,
            )
        relationships += '\n    ":Cube" -> "%s"' % coord_label

        # Are there any relationships to data dimensions?
        dims = cube.coord_dims(coord)
        for dim in dims:
            relationships_association += (
                '\n    "%s" -> "CubeDimension_%s":w' % (coord_label, dim)
            )

    dimension_nodes += """
    }
    """

    coord_nodes += """
    }
    """

    coord_system_nodes += """
    }
    """

    # return a string pulling everything together
    template = """
digraph CubeGraph{

    rankdir = "LR"
    fontname = "Bitstream Vera Sans"
    fontsize = 8

    node [
        fontname = "Bitstream Vera Sans"
        fontsize = 8
        shape = "record"
    ]

#   Nodes
%(cube_node)s
    %(dimension_nodes)s
    %(coord_nodes)s
    %(coord_sys_nodes)s
    edge [
        arrowhead = "normal"
    ]

#   RELATIONSHIPS

#   Containment
    %(relationships)s
    edge [
        style="dashed"
        arrowhead = "onormal"
    ]

#   Association
    %(associations)s
}
    """
    cube_attributes = list(
        sorted(cube.attributes.items(), key=lambda item: item[0])
    )
    cube_node = _dot_node(_GRAPH_INDENT, ":Cube", "Cube", cube_attributes)
    res_string = template % {
        "cube_node": cube_node,
        "dimension_nodes": dimension_nodes,
        "coord_nodes": coord_nodes,
        "coord_sys_nodes": coord_system_nodes,
        "relationships": relationships,
        "associations": relationships_association,
    }
    return res_string


def _coord_text(label, coord):
    """
    Returns a string containing the dot representation for a single coordinate node.

    Args:

    * label
        The dot ID of the coordinate node.
    * coord
        The coordinate to convert.

    """
    # Which bits to write?
    # Note: This is is not very OO but we are achieving a separation of DOT from cdm by doing this.
    if isinstance(coord, iris.coords.DimCoord):
        _dot_attrs = ("standard_name", "long_name", "units", "circular")
    elif isinstance(coord, iris.coords.AuxCoord):
        _dot_attrs = ("standard_name", "long_name", "units")
    else:
        raise ValueError("Unhandled coordinate type: " + str(type(coord)))
    attrs = [(name, getattr(coord, name)) for name in _dot_attrs]

    if coord.attributes:
        custom_attrs = sorted(
            coord.attributes.items(), key=lambda item: item[0]
        )
        attrs.extend(custom_attrs)

    node = _dot_node(_SUBGRAPH_INDENT, label, coord.__class__.__name__, attrs)
    return node


def _coord_system_text(cs, uid):
    """
    Returns a string containing the dot representation for a single coordinate system node.

    Args:

    * cs
        The coordinate system to convert.
    * uid
        The uid allows/distinguishes non-identical CoordSystems of the same type.

    """
    attrs = []
    for k, v in cs.__dict__.items():
        if isinstance(v, iris.cube.Cube):
            attrs.append((k, "defined"))
        else:
            attrs.append((k, v))

    attrs.sort(key=lambda attr: attr[0])

    label = "CoordSystem_%s_%s" % (cs.__class__.__name__, uid)
    node = _dot_node(_SUBGRAPH_INDENT, label, cs.__class__.__name__, attrs)
    return node


def _dot_node(indent, id, name, attributes):
    """
    Returns a string containing the dot representation for a single node.

    Args:

     * id
        The ID of the node.
     * name
        The visual name of the node.
     * attributes
        An iterable of (name, value) attribute pairs.

    """
    attributes = r"\n".join("%s: %s" % item for item in attributes)
    template = """%(indent)s"%(id)s" [
%(indent)s    label = "%(name)s|%(attributes)s"
%(indent)s]
"""
    node = template % {
        "id": id,
        "name": name,
        "attributes": attributes,
        "indent": indent,
    }
    return node



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for :func:`iris.util.new_axis`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import iris
from iris.util import new_axis


class Test(tests.IrisTest):
    @tests.skip_data
    def test_lazy_data(self):
        filename = tests.get_data_path(("PP", "globClim1", "theta.pp"))
        cube = iris.load_cube(filename)
        new_cube = new_axis(cube, "time")
        self.assertTrue(cube.has_lazy_data())
        self.assertTrue(new_cube.has_lazy_data())
        self.assertEqual(new_cube.shape, (1,) + cube.shape)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.abf.ABFField` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from iris.fileformats.abf import ABFField


class MethodCounter:
    def __init__(self, method_name):
        self.method_name = method_name
        self.count = 0

    def __enter__(self):
        self.orig_method = getattr(ABFField, self.method_name)

        def new_method(*args, **kwargs):
            self.count += 1
            self.orig_method(*args, **kwargs)

        setattr(ABFField, self.method_name, new_method)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        setattr(ABFField, self.method_name, self.orig_method)
        return False


class Test_data(tests.IrisTest):
    def test_single_read(self):
        path = "0000000000000000jan00000"
        field = ABFField(path)

        with mock.patch("iris.fileformats.abf.np.fromfile") as fromfile:
            with MethodCounter("__getattr__") as getattr:
                with MethodCounter("_read") as read:
                    field.data

        fromfile.assert_called_once_with(path, dtype=">u1")
        self.assertEqual(getattr.count, 1)
        self.assertEqual(read.count, 1)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Cube functions for iteration in step.

"""

from collections.abc import Iterator
import itertools
import warnings

import numpy as np


__all__ = ["izip"]


def izip(*cubes, **kwargs):
    """
    Return an iterator for iterating over a collection of cubes in step.

    If the input cubes have dimensions for which there are no common
    coordinates, those dimensions will be treated as orthogonal. The
    resulting iterator will step through combinations of the associated
    coordinates.

    Args:

    * cubes (:class:`iris.cube.Cube`):
        One or more :class:`iris.cube.Cube` instances over which to iterate in
        step. Each cube should be provided as a separate argument e.g.
        ``iris.iterate.izip(cube_a, cube_b, cube_c, ...)``.

    Kwargs:

    * coords (string, coord or a list of strings/coords):
        Coordinate names/coordinates of the desired subcubes (i.e. those
        that are not iterated over). They must all be orthogonal (i.e. point
        to different dimensions).
    * ordered (Boolean):
        If True (default), the order of the coordinates in the resulting
        subcubes will match the order of the coordinates in the coords
        keyword argument. If False, the order of the coordinates will
        be preserved and will match that of the input cubes.

    Returns:
        An iterator over a collection of tuples that contain the resulting
        subcubes.

    For example:
        >>> e_content, e_density = iris.load_cubes(
        ...     iris.sample_data_path('space_weather.nc'),
        ...     ['total electron content', 'electron density'])
        >>> for tslice, hslice in iris.iterate.izip(e_content, e_density,
        ...                                         coords=['grid_latitude',
        ...                                                 'grid_longitude']):
        ...    pass

    """
    if not cubes:
        raise TypeError("Expected one or more cubes.")

    ordered = kwargs.get("ordered", True)
    if not isinstance(ordered, bool):
        raise TypeError("Expected bool ordered parameter, got %r" % ordered)

    # Convert any coordinate names to coordinates (and ensure each cube has
    # requested slice coords).
    coords_to_slice = kwargs.get("coords")
    coords_by_cube = []

    for cube in cubes:
        if coords_to_slice is None or not coords_to_slice:
            coords_by_cube.append([])
        else:
            coords_by_cube.append(cube._as_list_of_coords(coords_to_slice))

    # For each input cube, generate the union of all describing dimensions for
    # the resulting subcube.
    requested_dims_by_cube = []
    for coords, cube in zip(coords_by_cube, cubes):
        requested_dims = set()
        for coord in coords:
            requested_dims.update(cube.coord_dims(coord))

        # Make sure this cube has no shared dimensions between the requested
        # coords.
        if len(requested_dims) != sum(
            (len(cube.coord_dims(coord)) for coord in coords)
        ):
            msg = (
                "The requested coordinates (%r) of cube (%r) are not "
                "orthogonal." % ([coord.name() for coord in coords], cube)
            )
            raise ValueError(msg)

        requested_dims_by_cube.append(requested_dims)

    # Checks on coordinates you are going to iterate over.
    # Create a list of sets (one set per cube), with each set containing the
    # dimensioned coordinates that will be iterated over (i.e exclude slice
    # coords).
    dimensioned_iter_coords_by_cube = []
    for requested_dims, cube in zip(requested_dims_by_cube, cubes):
        dimensioned_iter_coords = set()
        # Loop over dimensioned coords in each cube.
        for dim in range(len(cube.shape)):
            if dim not in requested_dims:
                dimensioned_iter_coords.update(
                    cube.coords(contains_dimension=dim)
                )
        dimensioned_iter_coords_by_cube.append(dimensioned_iter_coords)

    # Check for multidimensional coords - current implementation cannot
    # iterate over multidimensional coords.
    # for dimensioned_iter_coords in dimensioned_iter_coords_by_cube:
    #    for coord in dimensioned_iter_coords:
    #        if coord.ndim > 1:
    #            raise iris.exceptions.CoordinateMultiDimError(coord)

    # Iterate through all the possible pairs of cubes to compare dimensioned
    # coordinates.
    pairs_iter = itertools.combinations(dimensioned_iter_coords_by_cube, 2)
    for dimensioned_iter_coords_a, dimensioned_iter_coords_b in pairs_iter:
        coords_by_def_a = set(
            _CoordWrapper(coord) for coord in dimensioned_iter_coords_a
        )
        coords_by_def_b = set(
            _CoordWrapper(coord) for coord in dimensioned_iter_coords_b
        )

        # Check that the dimensioned coords that are common across the cubes
        # (i.e. have same definition/metadata) have the same shape. If this is
        # not the case it makes no sense to iterate through the coordinate in
        # step and an exception is raised.
        common = coords_by_def_a & coords_by_def_b
        for definition_coord in common:
            # Extract matching coord from dimensioned_iter_coords_a and
            # dimensioned_iter_coords_b to access shape.
            coord_a = next(
                coord
                for coord in dimensioned_iter_coords_a
                if definition_coord == coord
            )
            coord_b = next(
                coord
                for coord in dimensioned_iter_coords_b
                if definition_coord == coord
            )
            if coord_a.shape != coord_b.shape:
                raise ValueError(
                    "Shape of common dimensioned coordinate '%s' "
                    "does not match across all cubes. Unable "
                    "to iterate over this coordinate in "
                    "step." % coord_a.name()
                )
            if coord_a != coord_b:
                warnings.warn(
                    "Iterating over coordinate '%s' in step whose "
                    "definitions match but whose values "
                    "differ." % coord_a.name()
                )

    return _ZipSlicesIterator(
        cubes, requested_dims_by_cube, ordered, coords_by_cube
    )


class _ZipSlicesIterator(Iterator):
    """
    Extension to _SlicesIterator (see cube.py) to support iteration over a
    collection of cubes in step.

    """

    def __init__(self, cubes, requested_dims_by_cube, ordered, coords_by_cube):
        self._cubes = cubes
        self._requested_dims_by_cube = requested_dims_by_cube
        self._ordered = ordered
        self._coords_by_cube = coords_by_cube

        # Check that the requested_dims_by_cube and coords_by_cube lists are
        # the same length as cubes so it is feasible that there is a 1-1
        # mapping of values (itertool.izip won't catch this).
        if len(requested_dims_by_cube) != len(cubes):
            raise ValueError(
                "requested_dims_by_cube parameter is not the same"
                " length as cubes."
            )
        if len(coords_by_cube) != len(cubes):
            raise ValueError(
                "coords_by_cube parameter is not the same length " "as cubes."
            )

        # Create an all encompassing dims_index called master_dims_index that
        # is iterated over (using np.ndindex) and from which the indices of the
        # subcubes can be extracted using offsets i.e. position of the
        # associated coord in the master_dims_index.
        master_dimensioned_coord_list = []
        master_dims_index = []
        self._offsets_by_cube = []
        for requested_dims, cube in zip(requested_dims_by_cube, cubes):
            # Create a list of the shape of each cube, and set the dimensions
            # which have been requested to length 1.
            dims_index = list(cube.shape)
            for dim in requested_dims:
                dims_index[dim] = 1
            offsets = []
            # Loop over dimensions in each cube.
            for i in range(len(cube.shape)):
                # Obtain the coordinates for this dimension.
                cube_coords = cube.coords(dimensions=i)
                found = False
                # Loop over coords in this dimension (could be just one).
                for coord in cube_coords:
                    # Search for coord in master_dimensioned_coord_list.
                    for j, master_coords in enumerate(
                        master_dimensioned_coord_list
                    ):
                        # Use coord wrapper with desired equality
                        # functionality.
                        if _CoordWrapper(coord) in master_coords:
                            offsets.append(j)
                            found = True
                            break
                    if found:
                        break
                # If a coordinate with an equivalent definition (i.e. same
                # metadata) is not found in the master_dimensioned_coord_list,
                # add the coords assocaited with the dimension to the list,
                # add the size of the dimension to the master_dims_index and
                # store the offset.
                if not found:
                    master_dimensioned_coord_list.append(
                        set((_CoordWrapper(coord) for coord in cube_coords))
                    )
                    master_dims_index.append(dims_index[i])
                    offsets.append(len(master_dims_index) - 1)
            # Store the offsets for each cube so they can be used in
            # _ZipSlicesIterator.next().
            self._offsets_by_cube.append(offsets)

        # Let Numpy do some work in providing all of the permutations of our
        # data shape based on the combination of dimension sizes called
        # master_dims_index. This functionality is something like:
        # ndindex(2, 1, 3) -> [(0, 0, 0), (0, 0, 1), (0, 0, 2), (1, 0, 0),
        # (1, 0, 1), (1, 0, 2)]
        self._ndindex = np.ndindex(*master_dims_index)

    def __next__(self):
        # When self._ndindex runs out it will raise StopIteration for us.
        master_index_tuple = next(self._ndindex)

        subcubes = []
        for offsets, requested_dims, coords, cube in zip(
            self._offsets_by_cube,
            self._requested_dims_by_cube,
            self._coords_by_cube,
            self._cubes,
        ):
            # Extract the index_list for each cube from the master index using
            # the offsets and for each of the spanning dimensions requested,
            # replace the index_list value (will be a zero from np.ndindex())
            # with a spanning slice.
            index_list = [master_index_tuple[x] for x in offsets]
            for dim in requested_dims:
                index_list[dim] = slice(None, None)
            # Extract slices from the cube
            subcube = cube[tuple(index_list)]
            # Call transpose if necessary (taken from _SlicesIterator in
            # cube.py).
            if self._ordered is True:
                transpose_order = []
                for coord in coords:
                    transpose_order += sorted(subcube.coord_dims(coord))
                if transpose_order != list(range(subcube.ndim)):
                    subcube.transpose(transpose_order)
            subcubes.append(subcube)

        return tuple(subcubes)

    next = __next__


class _CoordWrapper:
    """
    Class for creating a coordinate wrapper that allows the use of an
    alternative equality function based on metadata rather than
    metadata + points/bounds.

    .. note::

        Uses a lightweight/incomplete implementation of the Decorator
        pattern.

    """

    def __init__(self, coord):
        self._coord = coord

    # Methods of contained class we need to expose/use.
    @property
    def metadata(self):
        return self._coord.metadata

    # Methods of contained class we want to override/customise.
    def __eq__(self, other):
        return self._coord.metadata == other.metadata

    # Force use of __eq__ for set operations.
    def __hash__(self):
        return 1



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :class:`iris.analysis._interpolation.RectilinearInterpolator`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import datetime

from iris._lazy_data import as_lazy_data
import numpy as np

import iris
import iris.coords
import iris.cube
import iris.exceptions
import iris.tests.stock as stock
from iris.analysis._interpolation import RectilinearInterpolator


LINEAR = "linear"
NEAREST = "nearest"

EXTRAPOLATE = "extrapolate"


class ThreeDimCube(tests.IrisTest):
    def setUp(self):
        cube = stock.simple_3d_w_multidim_coords()
        cube.add_aux_coord(
            iris.coords.DimCoord(np.arange(2), "height", units="1"), 0
        )
        cube.add_dim_coord(
            iris.coords.DimCoord(np.arange(3), "latitude", units="1"), 1
        )
        cube.add_dim_coord(
            iris.coords.DimCoord(np.arange(4), "longitude", units="1"), 2
        )
        self.data = np.arange(24).reshape(2, 3, 4).astype(np.float32)
        cube.data = self.data
        self.cube = cube


class Test___init__(ThreeDimCube):
    def test_properties(self):
        interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, EXTRAPOLATE
        )

        self.assertEqual(interpolator.method, LINEAR)
        self.assertEqual(interpolator.extrapolation_mode, EXTRAPOLATE)

        # Access to cube property of the RectilinearInterpolator instance.
        self.assertEqual(interpolator.cube, self.cube)

        # Access to the resulting coordinate which we are interpolating over.
        self.assertEqual(interpolator.coords, [self.cube.coord("latitude")])


class Test___init____validation(ThreeDimCube):
    def test_interpolator_overspecified(self):
        # Over specification by means of interpolating over two coordinates
        # mapped to the same dimension.
        msg = (
            "Coordinates repeat a data dimension - "
            "the interpolation would be over-specified"
        )
        with self.assertRaisesRegex(ValueError, msg):
            RectilinearInterpolator(
                self.cube, ["wibble", "height"], LINEAR, EXTRAPOLATE
            )

    def test_interpolator_overspecified_scalar(self):
        # Over specification by means of interpolating over one dimension
        # coordinate and a scalar coordinate (not mapped to a dimension).
        self.cube.add_aux_coord(
            iris.coords.AuxCoord(1, long_name="scalar"), None
        )

        msg = (
            "Coordinates repeat a data dimension - "
            "the interpolation would be over-specified"
        )
        with self.assertRaisesRegex(ValueError, msg):
            RectilinearInterpolator(
                self.cube, ["wibble", "scalar"], LINEAR, EXTRAPOLATE
            )

    def test_interpolate__decreasing(self):
        def check_expected():
            # Check a simple case is equivalent to extracting the first row.
            self.interpolator = RectilinearInterpolator(
                self.cube, ["latitude"], LINEAR, EXTRAPOLATE
            )
            expected = self.data[:, 0:1, :]
            result = self.interpolator([[0]])
            self.assertArrayEqual(result.data, expected)

        # Check with normal cube.
        check_expected()
        # Check same result from a cube inverted in the latitude dimension.
        self.cube = self.cube[:, ::-1]
        check_expected()

    def test_interpolate_non_monotonic(self):
        self.cube.add_aux_coord(
            iris.coords.AuxCoord([0, 3, 2], long_name="non-monotonic"), 1
        )
        msg = (
            "Cannot interpolate over the non-monotonic coordinate "
            "non-monotonic."
        )
        with self.assertRaisesRegex(ValueError, msg):
            RectilinearInterpolator(
                self.cube, ["non-monotonic"], LINEAR, EXTRAPOLATE
            )


class Test___call___1D(ThreeDimCube):
    def setUp(self):
        ThreeDimCube.setUp(self)
        self.interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, EXTRAPOLATE
        )

    def test_interpolate_bad_coord_name(self):
        with self.assertRaises(iris.exceptions.CoordinateNotFoundError):
            RectilinearInterpolator(
                self.cube, ["doesnt exist"], LINEAR, EXTRAPOLATE
            )

    def test_interpolate_data_single(self):
        # Single sample point.
        result = self.interpolator([[1.5]])
        expected = self.data[:, 1:, :].mean(axis=1).reshape(2, 1, 4)
        self.assertArrayEqual(result.data, expected)

        foo_res = result.coord("foo").points
        bar_res = result.coord("bar").points
        expected_foo = (
            self.cube[:, 1:, :].coord("foo").points.mean(axis=0).reshape(1, 4)
        )
        expected_bar = (
            self.cube[:, 1:, :].coord("bar").points.mean(axis=0).reshape(1, 4)
        )

        self.assertArrayEqual(foo_res, expected_foo)
        self.assertArrayEqual(bar_res, expected_bar)

    def test_interpolate_data_multiple(self):
        # Multiple sample points for a single coordinate (these points are not
        # interpolated).
        result = self.interpolator([[1, 2]])
        self.assertArrayEqual(result.data, self.data[:, 1:3, :])

        foo_res = result.coord("foo").points
        bar_res = result.coord("bar").points
        expected_foo = self.cube[:, 1:, :].coord("foo").points
        expected_bar = self.cube[:, 1:, :].coord("bar").points

        self.assertArrayEqual(foo_res, expected_foo)
        self.assertArrayEqual(bar_res, expected_bar)

    def test_interpolate_data_linear_extrapolation(self):
        # Sample point outside the coordinate range.
        result = self.interpolator([[-1]])
        expected = self.data[:, 0:1] - (self.data[:, 1:2] - self.data[:, 0:1])
        self.assertArrayEqual(result.data, expected)

    def _extrapolation_dtype(self, dtype):
        self.cube.data = self.cube.data.astype(dtype)
        interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, extrapolation_mode="nan"
        )
        result = interpolator([[-1]])
        self.assertTrue(np.all(np.isnan(result.data)))

    def test_extrapolation_nan_float32(self):
        # Ensure np.nan in a float32 array results.
        self._extrapolation_dtype(np.float32)

    def test_extrapolation_nan_float64(self):
        # Ensure np.nan in a float64 array results.
        self._extrapolation_dtype(np.float64)

    def test_interpolate_data_error_on_extrapolation(self):
        msg = "One of the requested xi is out of bounds in dimension 0"
        interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, extrapolation_mode="error"
        )
        with self.assertRaisesRegex(ValueError, msg):
            interpolator([[-1]])

    def test_interpolate_data_unsupported_extrapolation(self):
        msg = "Extrapolation mode 'unsupported' not supported"
        with self.assertRaisesRegex(ValueError, msg):
            RectilinearInterpolator(
                self.cube,
                ["latitude"],
                LINEAR,
                extrapolation_mode="unsupported",
            )

    def test_multi_points_array(self):
        # Providing a multidimensional sample points for a 1D interpolation.
        # i.e. points given for two coordinates where there are only one
        # specified.
        msg = "Expected sample points for 1 coordinates, got 2."
        with self.assertRaisesRegex(ValueError, msg):
            self.interpolator([[1, 2], [1]])

    def test_interpolate_data_dtype_casting(self):
        data = self.data.astype(int)
        self.cube.data = data
        self.interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, EXTRAPOLATE
        )
        result = self.interpolator([[0.125]])
        self.assertEqual(result.data.dtype, np.float64)

    def test_default_collapse_scalar(self):
        interpolator = RectilinearInterpolator(
            self.cube, ["wibble"], LINEAR, EXTRAPOLATE
        )
        result = interpolator([0])
        self.assertEqual(result.shape, (3, 4))

    def test_collapse_scalar(self):
        interpolator = RectilinearInterpolator(
            self.cube, ["wibble"], LINEAR, EXTRAPOLATE
        )
        result = interpolator([0], collapse_scalar=True)
        self.assertEqual(result.shape, (3, 4))

    def test_no_collapse_scalar(self):
        interpolator = RectilinearInterpolator(
            self.cube, ["wibble"], LINEAR, EXTRAPOLATE
        )
        result = interpolator([0], collapse_scalar=False)
        self.assertEqual(result.shape, (1, 3, 4))

    def test_unsorted_datadim_mapping(self):
        # Currently unsorted data dimension mapping is not supported as the
        # indexing is not yet clever enough to remap the interpolated
        # coordinates.
        self.cube.transpose((0, 2, 1))
        interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, EXTRAPOLATE
        )
        msg = "Currently only increasing data_dims is supported."
        with self.assertRaisesRegex(NotImplementedError, msg):
            interpolator([0])


class Test___call___1D_circular(ThreeDimCube):
    # Note: all these test data interpolation.
    def setUp(self):
        ThreeDimCube.setUp(self)
        self.cube.coord("longitude")._points = np.linspace(
            0, 360, 4, endpoint=False
        )
        self.cube.coord("longitude").circular = True
        self.cube.coord("longitude").units = "degrees"
        self.interpolator = RectilinearInterpolator(
            self.cube, ["longitude"], LINEAR, extrapolation_mode="nan"
        )
        self.cube_reverselons = self.cube[:, :, ::-1]
        self.interpolator_reverselons = RectilinearInterpolator(
            self.cube_reverselons,
            ["longitude"],
            LINEAR,
            extrapolation_mode="nan",
        )

        self.testpoints_fully_wrapped = ([[180, 270]], [[-180, -90]])
        self.testpoints_partially_wrapped = ([[180, 90]], [[-180, 90]])
        self.testpoints_fully_wrapped_twice = (
            [np.linspace(-360, 360, 100)],
            [(np.linspace(-360, 360, 100) + 360) % 360],
        )

    def test_fully_wrapped(self):
        points, points_wrapped = self.testpoints_fully_wrapped
        expected = self.interpolator(points)
        result = self.interpolator(points_wrapped)
        self.assertArrayEqual(expected.data, result.data)

    def test_fully_wrapped_reversed_mainpoints(self):
        points, _ = self.testpoints_fully_wrapped
        expected = self.interpolator(points)
        result = self.interpolator_reverselons(points)
        self.assertArrayEqual(expected.data, result.data)

    def test_fully_wrapped_reversed_testpoints(self):
        _, points = self.testpoints_fully_wrapped
        expected = self.interpolator(points)
        result = self.interpolator_reverselons(points)
        self.assertArrayEqual(expected.data, result.data)

    def test_partially_wrapped(self):
        points, points_wrapped = self.testpoints_partially_wrapped
        expected = self.interpolator(points)
        result = self.interpolator(points_wrapped)
        self.assertArrayEqual(expected.data, result.data)

    def test_partially_wrapped_reversed_mainpoints(self):
        points, _ = self.testpoints_partially_wrapped
        expected = self.interpolator(points)
        result = self.interpolator_reverselons(points)
        self.assertArrayEqual(expected.data, result.data)

    def test_partially_wrapped_reversed_testpoints(self):
        points, _ = self.testpoints_partially_wrapped
        expected = self.interpolator(points)
        result = self.interpolator_reverselons(points)
        self.assertArrayEqual(expected.data, result.data)

    def test_fully_wrapped_twice(self):
        xs, xs_not_wrapped = self.testpoints_fully_wrapped_twice
        expected = self.interpolator(xs)
        result = self.interpolator(xs_not_wrapped)
        self.assertArrayEqual(expected.data, result.data)

    def test_fully_wrapped_twice_reversed_mainpoints(self):
        _, points = self.testpoints_fully_wrapped_twice
        expected = self.interpolator(points)
        result = self.interpolator_reverselons(points)
        self.assertArrayEqual(expected.data, result.data)

    def test_fully_wrapped_not_circular(self):
        cube = stock.lat_lon_cube()
        new_long = cube.coord("longitude").copy(
            cube.coord("longitude").points + 710
        )
        cube.remove_coord("longitude")
        cube.add_dim_coord(new_long, 1)

        interpolator = RectilinearInterpolator(
            cube, ["longitude"], LINEAR, EXTRAPOLATE
        )
        res = interpolator([-10])
        self.assertArrayEqual(res.data, cube[:, 1].data)


class Test___call___1D_singlelendim(ThreeDimCube):
    def setUp(self):
        """
        thingness / (1)                     (wibble: 2; latitude: 1)
             Dimension coordinates:
                  wibble                           x            -
                  latitude                         -            x
             Auxiliary coordinates:
                  height                           x            -
                  bar                              -            x
                  foo                              -            x
             Scalar coordinates:
                  longitude: 0
        """
        ThreeDimCube.setUp(self)
        self.cube = self.cube[:, 0:1, 0]
        self.interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, EXTRAPOLATE
        )

    def test_interpolate_data_linear_extrapolation(self):
        # Linear extrapolation of a single valued element.
        result = self.interpolator([[1001]])
        self.assertArrayEqual(result.data, self.cube.data)

    def test_interpolate_data_nan_extrapolation(self):
        interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, extrapolation_mode="nan"
        )
        result = interpolator([[1001]])
        self.assertTrue(np.all(np.isnan(result.data)))

    def test_interpolate_data_nan_extrapolation_not_needed(self):
        # No extrapolation for a single length dimension.
        interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, extrapolation_mode="nan"
        )
        result = interpolator([[0]])
        self.assertArrayEqual(result.data, self.cube.data)


class Test___call___masked(tests.IrisTest):
    def setUp(self):
        self.cube = stock.simple_4d_with_hybrid_height()
        mask = np.isnan(self.cube.data)
        mask[::3, ::3] = True
        self.cube.data = np.ma.masked_array(self.cube.data, mask=mask)

    def test_orthogonal_cube(self):
        interpolator = RectilinearInterpolator(
            self.cube, ["grid_latitude"], LINEAR, EXTRAPOLATE
        )
        result_cube = interpolator([1])

        # Explicit mask comparison to ensure mask retention.
        # Masked value input
        self.assertTrue(self.cube.data.mask[0, 0, 0, 0])
        # Mask retention on output
        self.assertTrue(result_cube.data.mask[0, 0, 0])

        self.assertCML(
            result_cube,
            (
                "experimental",
                "analysis",
                "interpolate",
                "LinearInterpolator",
                "orthogonal_cube_with_factory.cml",
            ),
        )


class Test___call___2D(ThreeDimCube):
    def setUp(self):
        ThreeDimCube.setUp(self)
        self.interpolator = RectilinearInterpolator(
            self.cube, ["latitude", "longitude"], LINEAR, EXTRAPOLATE
        )

    def test_interpolate_data(self):
        result = self.interpolator([[1, 2], [2]])
        expected = self.data[:, 1:3, 2:3]
        self.assertArrayEqual(result.data, expected)

        index = (slice(None), slice(1, 3, 1), slice(2, 3, 1))
        for coord in self.cube.coords():
            coord_res = result.coord(coord).points
            coord_expected = self.cube[index].coord(coord).points

            self.assertArrayEqual(coord_res, coord_expected)

    def test_orthogonal_points(self):
        result = self.interpolator([[1, 2], [1, 2]])
        expected = self.data[:, 1:3, 1:3]
        self.assertArrayEqual(result.data, expected)

        index = (slice(None), slice(1, 3, 1), slice(1, 3, 1))
        for coord in self.cube.coords():
            coord_res = result.coord(coord).points
            coord_expected = self.cube[index].coord(coord).points

            self.assertArrayEqual(coord_res, coord_expected)

    def test_multi_dim_coord_interpolation(self):
        msg = "Interpolation coords must be 1-d for rectilinear interpolation."
        with self.assertRaisesRegex(ValueError, msg):
            interpolator = RectilinearInterpolator(
                self.cube, ["foo", "bar"], LINEAR, EXTRAPOLATE
            )
            interpolator([[15], [10]])


class Test___call___2D_non_contiguous(ThreeDimCube):
    def setUp(self):
        ThreeDimCube.setUp(self)
        coords = ["height", "longitude"]
        self.interpolator = RectilinearInterpolator(
            self.cube, coords, LINEAR, EXTRAPOLATE
        )

    def test_interpolate_data_multiple(self):
        result = self.interpolator([[1], [1, 2]])
        expected = self.data[1:2, :, 1:3]
        self.assertArrayEqual(result.data, expected)

        index = (slice(1, 2), slice(None), slice(1, 3, 1))
        for coord in self.cube.coords():
            coord_res = result.coord(coord).points
            coord_expected = self.cube[index].coord(coord).points

            self.assertArrayEqual(coord_res, coord_expected)

    def test_orthogonal_cube(self):
        result_cube = self.interpolator(
            [np.int64([0, 1, 1]), np.int32([0, 1])]
        )
        result_path = (
            "experimental",
            "analysis",
            "interpolate",
            "LinearInterpolator",
            "basic_orthogonal_cube.cml",
        )
        self.assertCMLApproxData(result_cube, result_path)
        self.assertEqual(result_cube.coord("longitude").dtype, np.int32)
        self.assertEqual(result_cube.coord("height").dtype, np.int64)

    def test_orthogonal_cube_squash(self):
        result_cube = self.interpolator([np.int64(0), np.int32([0, 1])])
        result_path = (
            "experimental",
            "analysis",
            "interpolate",
            "LinearInterpolator",
            "orthogonal_cube_1d_squashed.cml",
        )
        self.assertCMLApproxData(result_cube, result_path)
        self.assertEqual(result_cube.coord("longitude").dtype, np.int32)
        self.assertEqual(result_cube.coord("height").dtype, np.int64)

        non_collapsed_cube = self.interpolator(
            [[np.int64(0)], np.int32([0, 1])], collapse_scalar=False
        )
        result_path = (
            "experimental",
            "analysis",
            "interpolate",
            "LinearInterpolator",
            "orthogonal_cube_1d_squashed_2.cml",
        )
        self.assertCML(non_collapsed_cube[0, ...], result_path)
        self.assertCML(result_cube, result_path)
        self.assertEqual(result_cube, non_collapsed_cube[0, ...])


class Test___call___lazy_data(ThreeDimCube):
    def test_src_cube_data_loaded(self):
        # RectilinearInterpolator operates using a snapshot of the source cube.
        # If the source cube has lazy data when the interpolator is
        # instantiated we want to make sure the source cube's data is
        # loaded as a consequence of interpolation to avoid the risk
        # of loading it again and again.

        # Modify self.cube to have lazy data.
        self.cube.data = as_lazy_data(self.data)
        self.assertTrue(self.cube.has_lazy_data())

        # Perform interpolation and check the data has been loaded.
        interpolator = RectilinearInterpolator(
            self.cube, ["latitude"], LINEAR, EXTRAPOLATE
        )
        interpolator([[1.5]])
        self.assertFalse(self.cube.has_lazy_data())


class Test___call___time(tests.IrisTest):
    def interpolator(self, method=LINEAR):
        data = np.arange(12).reshape(4, 3)
        cube = iris.cube.Cube(data)
        time_coord = iris.coords.DimCoord(
            np.arange(0.0, 48.0, 12.0), "time", units="hours since epoch"
        )
        height_coord = iris.coords.DimCoord(
            np.arange(3), "altitude", units="m"
        )
        cube.add_dim_coord(time_coord, 0)
        cube.add_dim_coord(height_coord, 1)
        return RectilinearInterpolator(cube, ["time"], method, EXTRAPOLATE)

    def test_number_at_existing_value(self):
        interpolator = self.interpolator()
        result = interpolator([12])
        self.assertArrayEqual(result.data, [3, 4, 5])

    def test_datetime_at_existing_value(self):
        interpolator = self.interpolator()
        result = interpolator([datetime.datetime(1970, 1, 1, 12)])
        self.assertArrayEqual(result.data, [3, 4, 5])

    def test_datetime_between_existing_values(self):
        interpolator = self.interpolator()
        result = interpolator([datetime.datetime(1970, 1, 1, 18)])
        self.assertArrayEqual(result.data, [4.5, 5.5, 6.5])

    def test_mixed_numbers_and_datetimes(self):
        interpolator = self.interpolator()
        result = interpolator(
            [
                (
                    12,
                    datetime.datetime(1970, 1, 1, 18),
                    datetime.datetime(1970, 1, 2, 0),
                    26,
                )
            ]
        )
        self.assertEqual(result.coord("time").points.dtype, float)
        self.assertArrayEqual(
            result.data,
            [[3, 4, 5], [4.5, 5.5, 6.5], [6, 7, 8], [6.5, 7.5, 8.5]],
        )

    def test_mixed_numbers_and_datetimes_nearest(self):
        interpolator = self.interpolator(NEAREST)
        result = interpolator(
            [
                (
                    12,
                    datetime.datetime(1970, 1, 1, 18),
                    datetime.datetime(1970, 1, 2, 0),
                    26,
                )
            ]
        )
        self.assertEqual(result.coord("time").points.dtype, float)
        self.assertArrayEqual(
            result.data, [[3, 4, 5], [3, 4, 5], [6, 7, 8], [6, 7, 8]]
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

from iris.std_names import STD_NAMES


class TestStandardNames(tests.IrisTest):
    """
    standard_names.py is a machine generated file which contains a single dictionary
    called STD_NAMES
    """

    longMessage = True

    def test_standard_names(self):
        # Check we have a dict
        self.assertIsInstance(STD_NAMES, dict)

        keyset = set(STD_NAMES)

        # Check for some known standard names
        valid_nameset = set(["air_density", "northward_wind", "wind_speed"])
        self.assertTrue(valid_nameset.issubset(keyset), "Known standard name missing from STD_NAMES")

        # Check for some invalid standard names
        invalid_nameset = set(["invalid_air_density", "invalid_northward_wind",
                               "invalid_wind_speed",
                               "stratiform_snowfall_rate"])
        self.assertSetEqual(invalid_nameset - keyset, invalid_nameset,
                            "\nInvalid standard name(s) present in STD_NAMES")


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Definitions of derived coordinates.

"""

from abc import ABCMeta, abstractmethod
import warnings

import cf_units
import dask.array as da
import numpy as np

from iris.common import (
    CFVariableMixin,
    CoordMetadata,
    metadata_manager_factory,
)
import iris.coords


class AuxCoordFactory(CFVariableMixin, metaclass=ABCMeta):
    """
    Represents a "factory" which can manufacture an additional auxiliary
    coordinate on demand, by combining the values of other coordinates.

    Each concrete subclass represents a specific formula for deriving
    values from other coordinates.

    The `standard_name`, `long_name`, `var_name`, `units`, `attributes` and
    `coord_system` of the factory are used to set the corresponding
    properties of the resulting auxiliary coordinates.

    """

    def __init__(self):
        # Configure the metadata manager.
        if not hasattr(self, "_metadata_manager"):
            self._metadata_manager = metadata_manager_factory(CoordMetadata)

        #: Descriptive name of the coordinate made by the factory
        self.long_name = None

        #: netCDF variable name for the coordinate made by the factory
        self.var_name = None

        self.coord_system = None
        # See the climatological property getter.
        self._metadata_manager.climatological = False

    @property
    def coord_system(self):
        """
        The coordinate-system (if any) of the coordinate made by the factory.

        """
        return self._metadata_manager.coord_system

    @coord_system.setter
    def coord_system(self, value):
        self._metadata_manager.coord_system = value

    @property
    def climatological(self):
        """
        Always returns False, as a factory itself can never have points/bounds
        and therefore can never be climatological by definition.

        """
        return self._metadata_manager.climatological

    @property
    @abstractmethod
    def dependencies(self):
        """
        Returns a dictionary mapping from constructor argument names to
        the corresponding coordinates.

        """

    @abstractmethod
    def make_coord(self, coord_dims_func):
        """
        Returns a new :class:`iris.coords.AuxCoord` as defined by this
        factory.

        Args:

        * coord_dims_func:
            A callable which can return the list of dimensions relevant
            to a given coordinate.
            See :meth:`iris.cube.Cube.coord_dims()`.

        """

    @abstractmethod
    def update(self, old_coord, new_coord=None):
        """
        Notifies the factory of a removal/replacement of a dependency.

        Args:

        * old_coord:
            The dependency coordinate to be removed/replaced.
        * new_coord:
            If None, the dependency using old_coord is removed, otherwise
            the dependency is updated to use new_coord.

        """

    def __repr__(self):
        def arg_text(item):
            key, coord = item
            return "{}={}".format(key, str(coord and repr(coord.name())))

        items = sorted(self.dependencies.items(), key=lambda item: item[0])
        args = map(arg_text, items)
        return "<{}({})>".format(type(self).__name__, ", ".join(args))

    def derived_dims(self, coord_dims_func):
        """
        Returns the cube dimensions for the derived coordinate.

        Args:

        * coord_dims_func:
            A callable which can return the list of dimensions relevant
            to a given coordinate.
            See :meth:`iris.cube.Cube.coord_dims()`.

        Returns:

            A sorted list of cube dimension numbers.

        """
        # Which dimensions are relevant?
        # e.g. If sigma -> [1] and orog -> [2, 3] then result = [1, 2, 3]
        derived_dims = set()
        for coord in self.dependencies.values():
            if coord:
                derived_dims.update(coord_dims_func(coord))

        # Apply a fixed order so we know how to map dependency dims to
        # our own dims (and so the Cube can map them to Cube dims).
        derived_dims = tuple(sorted(derived_dims))
        return derived_dims

    def updated(self, new_coord_mapping):
        """
        Creates a new instance of this factory where the dependencies
        are replaced according to the given mapping.

        Args:

        * new_coord_mapping:
            A dictionary mapping from the object IDs potentially used
            by this factory, to the coordinate objects that should be
            used instead.

        """
        new_dependencies = {}
        for key, coord in self.dependencies.items():
            if coord:
                coord = new_coord_mapping[id(coord)]
            new_dependencies[key] = coord
        return type(self)(**new_dependencies)

    def xml_element(self, doc):
        """
        Returns a DOM element describing this coordinate factory.

        """
        element = doc.createElement("coordFactory")
        for key, coord in self.dependencies.items():
            element.setAttribute(key, coord._xml_id())
        element.appendChild(self.make_coord().xml_element(doc))
        return element

    def _dependency_dims(self, coord_dims_func):
        dependency_dims = {}
        for key, coord in self.dependencies.items():
            if coord:
                dependency_dims[key] = coord_dims_func(coord)
        return dependency_dims

    @staticmethod
    def _nd_bounds(coord, dims, ndim):
        """
        Return a lazy bounds array for a dependency coordinate, 'coord'.

        The result is aligned to the first 'ndim' cube dimensions, and
        expanded to the full ('ndim'+1)-dimensional shape.

        The value of 'ndim' must be >= the highest cube dimension of the
        dependency coordinate.

        The extra final result dimension ('ndim'-th) is the bounds dimension.

        Example:
            coord.shape == (70,)
            coord.nbounds = 2
            dims == [3]
            ndim == 5
        results in:
            nd_bounds.shape == (1, 1, 1, 70, 1, 2)

        """
        # Transpose to be consistent with the Cube.
        sorted_pairs = sorted(enumerate(dims), key=lambda pair: pair[1])
        transpose_order = [pair[0] for pair in sorted_pairs] + [len(dims)]
        bounds = coord.lazy_bounds()
        if dims and transpose_order != list(range(len(dims))):
            bounds = bounds.transpose(transpose_order)

        # Figure out the n-dimensional shape.
        nd_shape = [1] * ndim + [coord.nbounds]
        for dim, size in zip(dims, coord.shape):
            nd_shape[dim] = size
        bounds = bounds.reshape(nd_shape)
        return bounds

    @staticmethod
    def _nd_points(coord, dims, ndim):
        """
        Return a lazy points array for a dependency coordinate, 'coord'.

        The result is aligned to the first 'ndim' cube dimensions, and
        expanded to the full 'ndim'-dimensional shape.

        The value of 'ndim' must be >= the highest cube dimension of the
        dependency coordinate.

        Example:
            coord.shape == (4, 3)
            dims == [3, 2]
            ndim == 5
        results in:
            nd_points.shape == (1, 1, 3, 4, 1)

        """
        # Transpose to be consistent with the Cube.
        sorted_pairs = sorted(enumerate(dims), key=lambda pair: pair[1])
        transpose_order = [pair[0] for pair in sorted_pairs]
        points = coord.lazy_points()
        if dims and transpose_order != list(range(len(dims))):
            points = points.transpose(transpose_order)

        # Expand dimensionality to be consistent with the Cube.
        if dims:
            keys = [None] * ndim
            for dim, size in zip(dims, coord.shape):
                keys[dim] = slice(None)
            points = points[tuple(keys)]
        else:
            # Scalar coordinates have one dimensional points despite
            # mapping to zero dimensions, so we only need to add N-1
            # new dimensions.
            keys = (None,) * (ndim - 1)
            points = points[keys]
        return points

    def _remap(self, dependency_dims, derived_dims):
        """
        Return a mapping from dependency names to coordinate points arrays.

        For dependencies that are present, the values are all expanded and
        aligned to the same dimensions, which is the full set of all the
        dependency dimensions.
        These non-missing values are all lazy arrays.
        Missing dependencies, however, are assigned a scalar value of 0.0.

        """
        if derived_dims:
            ndim = max(derived_dims) + 1
        else:
            ndim = 1

        nd_points_by_key = {}
        for key, coord in self.dependencies.items():
            if coord:
                # Get the points as consistent with the Cube.
                nd_points = self._nd_points(coord, dependency_dims[key], ndim)

                # Restrict to just the dimensions relevant to the
                # derived coord. NB. These are always in Cube-order, so
                # no transpose is needed.
                if derived_dims:
                    keys = tuple(
                        slice(None) if dim in derived_dims else 0
                        for dim in range(ndim)
                    )
                    nd_points = nd_points[keys]
            else:
                # If no coord, treat value as zero.
                # Use a float16 to provide `shape` attribute and avoid
                # promoting other arguments to a higher precision.
                nd_points = np.float16(0)

            nd_points_by_key[key] = nd_points
        return nd_points_by_key

    def _remap_with_bounds(self, dependency_dims, derived_dims):
        """
        Return a mapping from dependency names to coordinate bounds arrays.

        For dependencies that are present, the values are all expanded and
        aligned to the same dimensions, which is the full set of all the
        dependency dimensions, plus an extra bounds dimension.
        These non-missing values are all lazy arrays.
        Missing dependencies, however, are assigned a scalar value of 0.0.

        Where a dependency coordinate has no bounds, then the associated value
        is taken from its points array, but reshaped to have an extra bounds
        dimension of length 1.

        """
        if derived_dims:
            ndim = max(derived_dims) + 1
        else:
            ndim = 1

        nd_values_by_key = {}
        for key, coord in self.dependencies.items():
            if coord:
                # Get the bounds or points as consistent with the Cube.
                if coord.nbounds:
                    nd_values = self._nd_bounds(
                        coord, dependency_dims[key], ndim
                    )
                else:
                    nd_values = self._nd_points(
                        coord, dependency_dims[key], ndim
                    )

                # Restrict to just the dimensions relevant to the
                # derived coord. NB. These are always in Cube-order, so
                # no transpose is needed.
                shape = []
                for dim in derived_dims:
                    shape.append(nd_values.shape[dim])
                # Ensure the array always has at least one dimension to be
                # compatible with normal coordinates.
                if not derived_dims:
                    shape.append(1)
                # Add on the N-bounds dimension
                if coord.nbounds:
                    shape.append(nd_values.shape[-1])
                else:
                    # NB. For a non-bounded coordinate we still need an
                    # extra dimension to make the shape compatible, so
                    # we just add an extra 1.
                    shape.append(1)
                nd_values = nd_values.reshape(shape)
            else:
                # If no coord, treat value as zero.
                # Use a float16 to provide `shape` attribute and avoid
                # promoting other arguments to a higher precision.
                nd_values = np.float16(0)

            nd_values_by_key[key] = nd_values
        return nd_values_by_key


class HybridHeightFactory(AuxCoordFactory):
    """
    Defines a hybrid-height coordinate factory with the formula:
        z = a + b * orog

    """

    def __init__(self, delta=None, sigma=None, orography=None):
        """
        Creates a hybrid-height coordinate factory with the formula:
            z = a + b * orog

        At least one of `delta` or `orography` must be provided.

        Args:

        * delta: Coord
            The coordinate providing the `a` term.
        * sigma: Coord
            The coordinate providing the `b` term.
        * orography: Coord
            The coordinate providing the `orog` term.

        """
        # Configure the metadata manager.
        self._metadata_manager = metadata_manager_factory(CoordMetadata)
        super().__init__()

        if delta and delta.nbounds not in (0, 2):
            raise ValueError(
                "Invalid delta coordinate: must have either 0 or" " 2 bounds."
            )
        if sigma and sigma.nbounds not in (0, 2):
            raise ValueError(
                "Invalid sigma coordinate: must have either 0 or" " 2 bounds."
            )
        if orography and orography.nbounds:
            msg = (
                "Orography coordinate {!r} has bounds."
                " These will be disregarded.".format(orography.name())
            )
            warnings.warn(msg, UserWarning, stacklevel=2)

        self.delta = delta
        self.sigma = sigma
        self.orography = orography

        self.standard_name = "altitude"
        if delta is None and orography is None:
            emsg = (
                "Unable to determine units: no delta or orography "
                "available."
            )
            raise ValueError(emsg)
        if delta and orography and delta.units != orography.units:
            emsg = (
                "Incompatible units: delta and orography must have "
                "the same units."
            )
            raise ValueError(emsg)
        self.units = (delta and delta.units) or orography.units
        if not self.units.is_convertible("m"):
            emsg = (
                "Invalid units: delta and/or orography must be expressed "
                "in length units."
            )
            raise ValueError(emsg)
        self.attributes = {"positive": "up"}

    @property
    def dependencies(self):
        """
        Returns a dictionary mapping from constructor argument names to
        the corresponding coordinates.

        """
        return {
            "delta": self.delta,
            "sigma": self.sigma,
            "orography": self.orography,
        }

    def _derive(self, delta, sigma, orography):
        return delta + sigma * orography

    def make_coord(self, coord_dims_func):
        """
        Returns a new :class:`iris.coords.AuxCoord` as defined by this
        factory.

        Args:

        * coord_dims_func:
            A callable which can return the list of dimensions relevant
            to a given coordinate.
            See :meth:`iris.cube.Cube.coord_dims()`.

        """
        # Which dimensions are relevant?
        derived_dims = self.derived_dims(coord_dims_func)
        dependency_dims = self._dependency_dims(coord_dims_func)

        # Build the points array.
        nd_points_by_key = self._remap(dependency_dims, derived_dims)
        points = self._derive(
            nd_points_by_key["delta"],
            nd_points_by_key["sigma"],
            nd_points_by_key["orography"],
        )

        bounds = None
        if (self.delta and self.delta.nbounds) or (
            self.sigma and self.sigma.nbounds
        ):
            # Build the bounds array.
            nd_values_by_key = self._remap_with_bounds(
                dependency_dims, derived_dims
            )
            delta = nd_values_by_key["delta"]
            sigma = nd_values_by_key["sigma"]
            orography = nd_values_by_key["orography"]
            ok_bound_shapes = [(), (1,), (2,)]
            if delta.shape[-1:] not in ok_bound_shapes:
                raise ValueError("Invalid delta coordinate bounds.")
            if sigma.shape[-1:] not in ok_bound_shapes:
                raise ValueError("Invalid sigma coordinate bounds.")
            if orography.shape[-1:] not in [(), (1,)]:
                warnings.warn(
                    "Orography coordinate has bounds. "
                    "These are being disregarded.",
                    UserWarning,
                    stacklevel=2,
                )
                orography_pts = nd_points_by_key["orography"]
                bds_shape = list(orography_pts.shape) + [1]
                orography = orography_pts.reshape(bds_shape)

            bounds = self._derive(delta, sigma, orography)

        hybrid_height = iris.coords.AuxCoord(
            points,
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            units=self.units,
            bounds=bounds,
            attributes=self.attributes,
            coord_system=self.coord_system,
        )
        return hybrid_height

    def update(self, old_coord, new_coord=None):
        """
        Notifies the factory of the removal/replacement of a coordinate
        which might be a dependency.

        Args:

        * old_coord:
            The coordinate to be removed/replaced.
        * new_coord:
            If None, any dependency using old_coord is removed, otherwise
            any dependency using old_coord is updated to use new_coord.

        """
        if self.delta is old_coord:
            if new_coord and new_coord.nbounds not in (0, 2):
                raise ValueError(
                    "Invalid delta coordinate:"
                    " must have either 0 or 2 bounds."
                )
            self.delta = new_coord
        elif self.sigma is old_coord:
            if new_coord and new_coord.nbounds not in (0, 2):
                raise ValueError(
                    "Invalid sigma coordinate:"
                    " must have either 0 or 2 bounds."
                )
            self.sigma = new_coord
        elif self.orography is old_coord:
            if new_coord and new_coord.nbounds:
                msg = (
                    "Orography coordinate {!r} has bounds."
                    " These will be disregarded.".format(new_coord.name())
                )
                warnings.warn(msg, UserWarning, stacklevel=2)
            self.orography = new_coord


class HybridPressureFactory(AuxCoordFactory):
    """
    Defines a hybrid-pressure coordinate factory with the formula:
        p = ap + b * ps

    """

    def __init__(self, delta=None, sigma=None, surface_air_pressure=None):
        """
        Creates a hybrid-height coordinate factory with the formula:
            p = ap + b * ps

        At least one of `delta` or `surface_air_pressure` must be provided.

        Args:

        * delta: Coord
            The coordinate providing the `ap` term.
        * sigma: Coord
            The coordinate providing the `b` term.
        * surface_air_pressure: Coord
            The coordinate providing the `ps` term.

        """
        # Configure the metadata manager.
        self._metadata_manager = metadata_manager_factory(CoordMetadata)
        super().__init__()

        # Check that provided coords meet necessary conditions.
        self._check_dependencies(delta, sigma, surface_air_pressure)
        self.units = (delta and delta.units) or surface_air_pressure.units

        self.delta = delta
        self.sigma = sigma
        self.surface_air_pressure = surface_air_pressure

        self.standard_name = "air_pressure"
        self.attributes = {}

    @staticmethod
    def _check_dependencies(delta, sigma, surface_air_pressure):
        # Check for sufficient coordinates.
        if delta is None and (sigma is None or surface_air_pressure is None):
            msg = (
                "Unable to construct hybrid pressure coordinate factory "
                "due to insufficient source coordinates."
            )
            raise ValueError(msg)

        # Check bounds.
        if delta and delta.nbounds not in (0, 2):
            raise ValueError(
                "Invalid delta coordinate: must have either 0 or" " 2 bounds."
            )
        if sigma and sigma.nbounds not in (0, 2):
            raise ValueError(
                "Invalid sigma coordinate: must have either 0 or" " 2 bounds."
            )
        if surface_air_pressure and surface_air_pressure.nbounds:
            msg = (
                "Surface pressure coordinate {!r} has bounds. These will"
                " be disregarded.".format(surface_air_pressure.name())
            )
            warnings.warn(msg, UserWarning, stacklevel=2)

        # Check units.
        if sigma is not None and sigma.units.is_unknown():
            # Be graceful, and promote unknown to dimensionless units.
            sigma.units = cf_units.Unit("1")

        if sigma is not None and not sigma.units.is_dimensionless():
            raise ValueError("Invalid units: sigma must be dimensionless.")
        if (
            delta is not None
            and surface_air_pressure is not None
            and delta.units != surface_air_pressure.units
        ):
            msg = (
                "Incompatible units: delta and "
                "surface_air_pressure must have the same units."
            )
            raise ValueError(msg)

        if delta is not None:
            units = delta.units
        else:
            units = surface_air_pressure.units

        if not units.is_convertible("Pa"):
            msg = (
                "Invalid units: delta and "
                "surface_air_pressure must have units of pressure."
            )
            raise ValueError(msg)

    @property
    def dependencies(self):
        """
        Returns a dictionary mapping from constructor argument names to
        the corresponding coordinates.

        """
        return {
            "delta": self.delta,
            "sigma": self.sigma,
            "surface_air_pressure": self.surface_air_pressure,
        }

    def _derive(self, delta, sigma, surface_air_pressure):
        return delta + sigma * surface_air_pressure

    def make_coord(self, coord_dims_func):
        """
        Returns a new :class:`iris.coords.AuxCoord` as defined by this
        factory.

        Args:

        * coord_dims_func:
            A callable which can return the list of dimensions relevant
            to a given coordinate.
            See :meth:`iris.cube.Cube.coord_dims()`.

        """
        # Which dimensions are relevant?
        derived_dims = self.derived_dims(coord_dims_func)
        dependency_dims = self._dependency_dims(coord_dims_func)

        # Build the points array.
        nd_points_by_key = self._remap(dependency_dims, derived_dims)
        points = self._derive(
            nd_points_by_key["delta"],
            nd_points_by_key["sigma"],
            nd_points_by_key["surface_air_pressure"],
        )

        bounds = None
        if (self.delta and self.delta.nbounds) or (
            self.sigma and self.sigma.nbounds
        ):
            # Build the bounds array.
            nd_values_by_key = self._remap_with_bounds(
                dependency_dims, derived_dims
            )
            delta = nd_values_by_key["delta"]
            sigma = nd_values_by_key["sigma"]
            surface_air_pressure = nd_values_by_key["surface_air_pressure"]
            ok_bound_shapes = [(), (1,), (2,)]
            if delta.shape[-1:] not in ok_bound_shapes:
                raise ValueError("Invalid delta coordinate bounds.")
            if sigma.shape[-1:] not in ok_bound_shapes:
                raise ValueError("Invalid sigma coordinate bounds.")
            if surface_air_pressure.shape[-1:] not in [(), (1,)]:
                warnings.warn(
                    "Surface pressure coordinate has bounds. "
                    "These are being disregarded."
                )
                surface_air_pressure_pts = nd_points_by_key[
                    "surface_air_pressure"
                ]
                bds_shape = list(surface_air_pressure_pts.shape) + [1]
                surface_air_pressure = surface_air_pressure_pts.reshape(
                    bds_shape
                )

            bounds = self._derive(delta, sigma, surface_air_pressure)

        hybrid_pressure = iris.coords.AuxCoord(
            points,
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            units=self.units,
            bounds=bounds,
            attributes=self.attributes,
            coord_system=self.coord_system,
        )
        return hybrid_pressure

    def update(self, old_coord, new_coord=None):
        """
        Notifies the factory of the removal/replacement of a coordinate
        which might be a dependency.

        Args:

        * old_coord:
            The coordinate to be removed/replaced.
        * new_coord:
            If None, any dependency using old_coord is removed, otherwise
            any dependency using old_coord is updated to use new_coord.

        """
        new_dependencies = self.dependencies
        for name, coord in self.dependencies.items():
            if old_coord is coord:
                new_dependencies[name] = new_coord
                try:
                    self._check_dependencies(**new_dependencies)
                except ValueError as e:
                    msg = "Failed to update dependencies. " + str(e)
                    raise ValueError(msg)
                else:
                    setattr(self, name, new_coord)
                break


class OceanSigmaZFactory(AuxCoordFactory):
    """Defines an ocean sigma over z coordinate factory."""

    def __init__(
        self,
        sigma=None,
        eta=None,
        depth=None,
        depth_c=None,
        nsigma=None,
        zlev=None,
    ):
        """
        Creates an ocean sigma over z coordinate factory with the formula:

        if k < nsigma:
            z(n, k, j, i) = eta(n, j, i) + sigma(k) *
                             (min(depth_c, depth(j, i)) + eta(n, j, i))

        if k >= nsigma:
            z(n, k, j, i) = zlev(k)

        The `zlev` and 'nsigma' coordinates must be provided, and at least
        either `eta`, or 'sigma' and `depth` and `depth_c` coordinates.

        """
        # Configure the metadata manager.
        self._metadata_manager = metadata_manager_factory(CoordMetadata)
        super().__init__()

        # Check that provided coordinates meet necessary conditions.
        self._check_dependencies(sigma, eta, depth, depth_c, nsigma, zlev)
        self.units = zlev.units

        self.sigma = sigma
        self.eta = eta
        self.depth = depth
        self.depth_c = depth_c
        self.nsigma = nsigma
        self.zlev = zlev

        self.standard_name = "sea_surface_height_above_reference_ellipsoid"
        self.attributes = {"positive": "up"}

    @staticmethod
    def _check_dependencies(sigma, eta, depth, depth_c, nsigma, zlev):
        # Check for sufficient factory coordinates.
        if zlev is None:
            raise ValueError(
                "Unable to determine units: no zlev coordinate available."
            )
        if nsigma is None:
            raise ValueError("Missing nsigma coordinate.")

        if eta is None and (sigma is None or depth_c is None or depth is None):
            msg = (
                "Unable to construct ocean sigma over z coordinate "
                "factory due to insufficient source coordinates."
            )
            raise ValueError(msg)

        # Check bounds and shape.
        for coord, term in ((sigma, "sigma"), (zlev, "zlev")):
            if coord is not None and coord.nbounds not in (0, 2):
                msg = (
                    "Invalid {} coordinate {!r}: must have either "
                    "0 or 2 bounds.".format(term, coord.name())
                )
                raise ValueError(msg)

        if sigma and sigma.nbounds != zlev.nbounds:
            msg = (
                "The sigma coordinate {!r} and zlev coordinate {!r} "
                "must be equally bounded.".format(sigma.name(), zlev.name())
            )
            raise ValueError(msg)

        coords = (
            (eta, "eta"),
            (depth, "depth"),
            (depth_c, "depth_c"),
            (nsigma, "nsigma"),
        )
        for coord, term in coords:
            if coord is not None and coord.nbounds:
                msg = (
                    "The {} coordinate {!r} has bounds. "
                    "These are being disregarded.".format(term, coord.name())
                )
                warnings.warn(msg, UserWarning, stacklevel=2)

        for coord, term in ((depth_c, "depth_c"), (nsigma, "nsigma")):
            if coord is not None and coord.shape != (1,):
                msg = (
                    "Expected scalar {} coordinate {!r}: "
                    "got shape {!r}.".format(term, coord.name(), coord.shape)
                )
                raise ValueError(msg)

        # Check units.
        if not zlev.units.is_convertible("m"):
            msg = (
                "Invalid units: zlev coordinate {!r} "
                "must have units of distance.".format(zlev.name())
            )
            raise ValueError(msg)

        if sigma is not None and sigma.units.is_unknown():
            # Be graceful, and promote unknown to dimensionless units.
            sigma.units = cf_units.Unit("1")

        if sigma is not None and not sigma.units.is_dimensionless():
            msg = (
                "Invalid units: sigma coordinate {!r} "
                "must be dimensionless.".format(sigma.name())
            )
            raise ValueError(msg)

        coords = ((eta, "eta"), (depth_c, "depth_c"), (depth, "depth"))
        for coord, term in coords:
            if coord is not None and coord.units != zlev.units:
                msg = (
                    "Incompatible units: {} coordinate {!r} and zlev "
                    "coordinate {!r} must have "
                    "the same units.".format(term, coord.name(), zlev.name())
                )
                raise ValueError(msg)

    @property
    def dependencies(self):
        """
        Returns a dictionary mapping from constructor argument names to
        the corresponding coordinates.

        """
        return dict(
            sigma=self.sigma,
            eta=self.eta,
            depth=self.depth,
            depth_c=self.depth_c,
            nsigma=self.nsigma,
            zlev=self.zlev,
        )

    def _derive(
        self, sigma, eta, depth, depth_c, zlev, nsigma, coord_dims_func
    ):
        # Calculate the index of the 'z' dimension in the input arrays.
        # First find the cube 'z' dimension ...
        [cube_z_dim] = coord_dims_func(self.dependencies["zlev"])
        # ... then calculate the corresponding dependency dimension.
        derived_cubedims = self.derived_dims(coord_dims_func)
        z_dim = derived_cubedims.index(cube_z_dim)

        # Calculate the result shape as a combination of all the inputs.
        # Note: all the inputs have the same number of dimensions >= 1, except
        # for any missing dependencies, which have scalar values.
        allshapes = np.array(
            [
                el.shape
                for el in (sigma, eta, depth, depth_c, zlev)
                if el.ndim > 0
            ]
        )
        result_shape = list(np.max(allshapes, axis=0))
        ndims = len(result_shape)

        # Make a slice tuple to index the first nsigma z-levels.
        z_slices_nsigma = [slice(None)] * ndims
        z_slices_nsigma[z_dim] = slice(0, int(nsigma))
        z_slices_nsigma = tuple(z_slices_nsigma)
        # Make a slice tuple to index the remaining z-levels.
        z_slices_rest = [slice(None)] * ndims
        z_slices_rest[z_dim] = slice(int(nsigma), None)
        z_slices_rest = tuple(z_slices_rest)

        # Perform the ocean sigma over z coordinate nsigma slice.
        if eta.ndim:
            eta = eta[z_slices_nsigma]
        if sigma.ndim:
            sigma = sigma[z_slices_nsigma]
        if depth.ndim:
            depth = depth[z_slices_nsigma]
        # Note that, this performs a point-wise minimum.
        nsigma_levs = eta + sigma * (da.minimum(depth_c, depth) + eta)

        # Make a result-shaped lazy "ones" array for expanding partial results.
        # Note: for the 'chunks' arg, we try to use [1, 1, ... ny, nx].
        # This calculation could be assuming too much in some cases, as we
        # don't actually check the dimensions of our dependencies anywhere.
        result_chunks = result_shape
        if len(result_shape) > 1:
            result_chunks = [1] * len(result_shape)
            result_chunks[-2:] = result_shape[-2:]
        ones_full_result = da.ones(
            result_shape, chunks=result_chunks, dtype=zlev.dtype
        )

        # Expand nsigma_levs to its full required shape : needed as the
        # calculated result may have a fixed size of 1 in some dimensions.
        result_nsigma_levs = nsigma_levs * ones_full_result[z_slices_nsigma]

        # Likewise, expand zlev to its full required shape.
        result_rest_levs = (
            zlev[z_slices_rest] * ones_full_result[z_slices_rest]
        )

        # Combine nsigma and 'rest' levels for the final result.
        result = da.concatenate(
            [result_nsigma_levs, result_rest_levs], axis=z_dim
        )
        return result

    def make_coord(self, coord_dims_func):
        """
        Returns a new :class:`iris.coords.AuxCoord` as defined by this factory.

        Args:

        * coord_dims_func:
            A callable which can return the list of dimensions relevant
            to a given coordinate. See :meth:`iris.cube.Cube.coord_dims()`.

        """
        # Determine the relevant dimensions.
        derived_dims = self.derived_dims(coord_dims_func)
        dependency_dims = self._dependency_dims(coord_dims_func)

        # Build the points array.
        nd_points_by_key = self._remap(dependency_dims, derived_dims)

        [nsigma] = nd_points_by_key["nsigma"]
        points = self._derive(
            nd_points_by_key["sigma"],
            nd_points_by_key["eta"],
            nd_points_by_key["depth"],
            nd_points_by_key["depth_c"],
            nd_points_by_key["zlev"],
            nsigma,
            coord_dims_func,
        )

        bounds = None
        if self.zlev.nbounds or (self.sigma and self.sigma.nbounds):
            # Build the bounds array.
            nd_values_by_key = self._remap_with_bounds(
                dependency_dims, derived_dims
            )
            valid_shapes = [(), (1,), (2,)]
            for key in ("sigma", "zlev"):
                if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                    name = self.dependencies[key].name()
                    msg = "Invalid bounds for {} " "coordinate {!r}.".format(
                        key, name
                    )
                    raise ValueError(msg)
            valid_shapes.pop()
            for key in ("eta", "depth", "depth_c", "nsigma"):
                if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                    name = self.dependencies[key].name()
                    msg = (
                        "The {} coordinate {!r} has bounds. "
                        "These are being disregarded.".format(key, name)
                    )
                    warnings.warn(msg, UserWarning, stacklevel=2)
                    # Swap bounds with points.
                    bds_shape = list(nd_points_by_key[key].shape) + [1]
                    bounds = nd_points_by_key[key].reshape(bds_shape)
                    nd_values_by_key[key] = bounds

            bounds = self._derive(
                nd_values_by_key["sigma"],
                nd_values_by_key["eta"],
                nd_values_by_key["depth"],
                nd_values_by_key["depth_c"],
                nd_values_by_key["zlev"],
                nsigma,
                coord_dims_func,
            )

        coord = iris.coords.AuxCoord(
            points,
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            units=self.units,
            bounds=bounds,
            attributes=self.attributes,
            coord_system=self.coord_system,
        )
        return coord

    def update(self, old_coord, new_coord=None):
        """
        Notifies the factory of the removal/replacement of a coordinate
        which might be a dependency.

        Args:

        * old_coord:
            The coordinate to be removed/replaced.
        * new_coord:
            If None, any dependency using old_coord is removed, otherwise
            any dependency using old_coord is updated to use new_coord.

        """
        new_dependencies = self.dependencies
        for name, coord in self.dependencies.items():
            if old_coord is coord:
                new_dependencies[name] = new_coord
                try:
                    self._check_dependencies(**new_dependencies)
                except ValueError as e:
                    msg = "Failed to update dependencies. " + str(e)
                    raise ValueError(msg)
                else:
                    setattr(self, name, new_coord)
                break


class OceanSigmaFactory(AuxCoordFactory):
    """Defines an ocean sigma coordinate factory."""

    def __init__(self, sigma=None, eta=None, depth=None):
        """
        Creates an ocean sigma coordinate factory with the formula:

        z(n, k, j, i) = eta(n, j, i) + sigma(k) *
                        (depth(j, i) + eta(n, j, i))

        """
        # Configure the metadata manager.
        self._metadata_manager = metadata_manager_factory(CoordMetadata)
        super().__init__()

        # Check that provided coordinates meet necessary conditions.
        self._check_dependencies(sigma, eta, depth)
        self.units = depth.units

        self.sigma = sigma
        self.eta = eta
        self.depth = depth

        self.standard_name = "sea_surface_height_above_reference_ellipsoid"
        self.attributes = {"positive": "up"}

    @staticmethod
    def _check_dependencies(sigma, eta, depth):
        # Check for sufficient factory coordinates.
        if eta is None or sigma is None or depth is None:
            msg = (
                "Unable to construct ocean sigma coordinate "
                "factory due to insufficient source coordinates."
            )
            raise ValueError(msg)

        # Check bounds and shape.
        coord, term = (sigma, "sigma")
        if coord is not None and coord.nbounds not in (0, 2):
            msg = (
                "Invalid {} coordinate {!r}: must have either "
                "0 or 2 bounds.".format(term, coord.name())
            )
            raise ValueError(msg)

        coords = ((eta, "eta"), (depth, "depth"))
        for coord, term in coords:
            if coord is not None and coord.nbounds:
                msg = (
                    "The {} coordinate {!r} has bounds. "
                    "These are being disregarded.".format(term, coord.name())
                )
                warnings.warn(msg, UserWarning, stacklevel=2)

        # Check units.
        if sigma is not None and sigma.units.is_unknown():
            # Be graceful, and promote unknown to dimensionless units.
            sigma.units = cf_units.Unit("1")

        if sigma is not None and not sigma.units.is_dimensionless():
            msg = (
                "Invalid units: sigma coordinate {!r} "
                "must be dimensionless.".format(sigma.name())
            )
            raise ValueError(msg)

        coords = ((eta, "eta"), (depth, "depth"))
        for coord, term in coords:
            if coord is not None and coord.units != depth.units:
                msg = (
                    "Incompatible units: {} coordinate {!r} and depth "
                    "coordinate {!r} must have "
                    "the same units.".format(term, coord.name(), depth.name())
                )
                raise ValueError(msg)

    @property
    def dependencies(self):
        """
        Returns a dictionary mapping from constructor argument names to
        the corresponding coordinates.

        """
        return dict(sigma=self.sigma, eta=self.eta, depth=self.depth)

    def _derive(self, sigma, eta, depth):
        return eta + sigma * (depth + eta)

    def make_coord(self, coord_dims_func):
        """
        Returns a new :class:`iris.coords.AuxCoord` as defined by this factory.

        Args:

        * coord_dims_func:
            A callable which can return the list of dimensions relevant
            to a given coordinate. See :meth:`iris.cube.Cube.coord_dims()`.

        """
        # Determine the relevant dimensions.
        derived_dims = self.derived_dims(coord_dims_func)
        dependency_dims = self._dependency_dims(coord_dims_func)

        # Build the points array.
        nd_points_by_key = self._remap(dependency_dims, derived_dims)
        points = self._derive(
            nd_points_by_key["sigma"],
            nd_points_by_key["eta"],
            nd_points_by_key["depth"],
        )

        bounds = None
        if self.sigma and self.sigma.nbounds:
            # Build the bounds array.
            nd_values_by_key = self._remap_with_bounds(
                dependency_dims, derived_dims
            )
            valid_shapes = [(), (1,), (2,)]
            key = "sigma"
            if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                name = self.dependencies[key].name()
                msg = "Invalid bounds for {} " "coordinate {!r}.".format(
                    key, name
                )
                raise ValueError(msg)
            valid_shapes.pop()
            for key in ("eta", "depth"):
                if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                    name = self.dependencies[key].name()
                    msg = (
                        "The {} coordinate {!r} has bounds. "
                        "These are being disregarded.".format(key, name)
                    )
                    warnings.warn(msg, UserWarning, stacklevel=2)
                    # Swap bounds with points.
                    bds_shape = list(nd_points_by_key[key].shape) + [1]
                    bounds = nd_points_by_key[key].reshape(bds_shape)
                    nd_values_by_key[key] = bounds

            bounds = self._derive(
                nd_values_by_key["sigma"],
                nd_values_by_key["eta"],
                nd_values_by_key["depth"],
            )

        coord = iris.coords.AuxCoord(
            points,
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            units=self.units,
            bounds=bounds,
            attributes=self.attributes,
            coord_system=self.coord_system,
        )
        return coord

    def update(self, old_coord, new_coord=None):
        """
        Notifies the factory of the removal/replacement of a coordinate
        which might be a dependency.

        Args:

        * old_coord:
            The coordinate to be removed/replaced.
        * new_coord:
            If None, any dependency using old_coord is removed, otherwise
            any dependency using old_coord is updated to use new_coord.

        """
        new_dependencies = self.dependencies
        for name, coord in self.dependencies.items():
            if old_coord is coord:
                new_dependencies[name] = new_coord
                try:
                    self._check_dependencies(**new_dependencies)
                except ValueError as e:
                    msg = "Failed to update dependencies. " + str(e)
                    raise ValueError(msg)
                else:
                    setattr(self, name, new_coord)
                break


class OceanSg1Factory(AuxCoordFactory):
    """Defines an Ocean s-coordinate, generic form 1 factory."""

    def __init__(self, s=None, c=None, eta=None, depth=None, depth_c=None):
        """
        Creates an Ocean s-coordinate, generic form 1 factory with the formula:

        z(n,k,j,i) = S(k,j,i) + eta(n,j,i) * (1 + S(k,j,i) / depth(j,i))

        where:
            S(k,j,i) = depth_c * s(k) + (depth(j,i) - depth_c) * C(k)

        """
        # Configure the metadata manager.
        self._metadata_manager = metadata_manager_factory(CoordMetadata)
        super().__init__()

        # Check that provided coordinates meet necessary conditions.
        self._check_dependencies(s, c, eta, depth, depth_c)
        self.units = depth.units

        self.s = s
        self.c = c
        self.eta = eta
        self.depth = depth
        self.depth_c = depth_c

        self.standard_name = "sea_surface_height_above_reference_ellipsoid"
        self.attributes = {"positive": "up"}

    @staticmethod
    def _check_dependencies(s, c, eta, depth, depth_c):
        # Check for sufficient factory coordinates.
        if (
            eta is None
            or s is None
            or c is None
            or depth is None
            or depth_c is None
        ):
            msg = (
                "Unable to construct Ocean s-coordinate, generic form 1 "
                "factory due to insufficient source coordinates."
            )
            raise ValueError(msg)

        # Check bounds and shape.
        coords = ((s, "s"), (c, "c"))
        for coord, term in coords:
            if coord is not None and coord.nbounds not in (0, 2):
                msg = (
                    "Invalid {} coordinate {!r}: must have either "
                    "0 or 2 bounds.".format(term, coord.name())
                )
                raise ValueError(msg)

        if s and s.nbounds != c.nbounds:
            msg = (
                "The s coordinate {!r} and c coordinate {!r} "
                "must be equally bounded.".format(s.name(), c.name())
            )
            raise ValueError(msg)

        coords = ((eta, "eta"), (depth, "depth"))
        for coord, term in coords:
            if coord is not None and coord.nbounds:
                msg = (
                    "The {} coordinate {!r} has bounds. "
                    "These are being disregarded.".format(term, coord.name())
                )
                warnings.warn(msg, UserWarning, stacklevel=2)

        if depth_c is not None and depth_c.shape != (1,):
            msg = (
                "Expected scalar {} coordinate {!r}: "
                "got shape {!r}.".format(term, coord.name(), coord.shape)
            )
            raise ValueError(msg)

        # Check units.
        coords = ((s, "s"), (c, "c"))
        for coord, term in coords:
            if coord is not None and coord.units.is_unknown():
                # Be graceful, and promote unknown to dimensionless units.
                coord.units = cf_units.Unit("1")

            if coord is not None and not coord.units.is_dimensionless():
                msg = (
                    "Invalid units: {} coordinate {!r} "
                    "must be dimensionless.".format(term, coord.name())
                )
                raise ValueError(msg)

        coords = ((eta, "eta"), (depth, "depth"), (depth_c, "depth_c"))
        for coord, term in coords:
            if coord is not None and coord.units != depth.units:
                msg = (
                    "Incompatible units: {} coordinate {!r} and depth "
                    "coordinate {!r} must have "
                    "the same units.".format(term, coord.name(), depth.name())
                )
                raise ValueError(msg)

    @property
    def dependencies(self):
        """
        Returns a dictionary mapping from constructor argument names to
        the corresponding coordinates.

        """
        return dict(
            s=self.s,
            c=self.c,
            eta=self.eta,
            depth=self.depth,
            depth_c=self.depth_c,
        )

    def _derive(self, s, c, eta, depth, depth_c):
        S = depth_c * s + (depth - depth_c) * c
        return S + eta * (1 + S / depth)

    def make_coord(self, coord_dims_func):
        """
        Returns a new :class:`iris.coords.AuxCoord` as defined by this factory.

        Args:

        * coord_dims_func:
            A callable which can return the list of dimensions relevant
            to a given coordinate. See :meth:`iris.cube.Cube.coord_dims()`.

        """
        # Determine the relevant dimensions.
        derived_dims = self.derived_dims(coord_dims_func)
        dependency_dims = self._dependency_dims(coord_dims_func)

        # Build the points array.
        nd_points_by_key = self._remap(dependency_dims, derived_dims)
        points = self._derive(
            nd_points_by_key["s"],
            nd_points_by_key["c"],
            nd_points_by_key["eta"],
            nd_points_by_key["depth"],
            nd_points_by_key["depth_c"],
        )

        bounds = None
        if self.s.nbounds or (self.c and self.c.nbounds):
            # Build the bounds array.
            nd_values_by_key = self._remap_with_bounds(
                dependency_dims, derived_dims
            )
            valid_shapes = [(), (1,), (2,)]
            key = "s"
            if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                name = self.dependencies[key].name()
                msg = "Invalid bounds for {} " "coordinate {!r}.".format(
                    key, name
                )
                raise ValueError(msg)
            valid_shapes.pop()
            for key in ("eta", "depth", "depth_c"):
                if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                    name = self.dependencies[key].name()
                    msg = (
                        "The {} coordinate {!r} has bounds. "
                        "These are being disregarded.".format(key, name)
                    )
                    warnings.warn(msg, UserWarning, stacklevel=2)
                    # Swap bounds with points.
                    bds_shape = list(nd_points_by_key[key].shape) + [1]
                    bounds = nd_points_by_key[key].reshape(bds_shape)
                    nd_values_by_key[key] = bounds

            bounds = self._derive(
                nd_values_by_key["s"],
                nd_values_by_key["c"],
                nd_values_by_key["eta"],
                nd_values_by_key["depth"],
                nd_values_by_key["depth_c"],
            )

        coord = iris.coords.AuxCoord(
            points,
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            units=self.units,
            bounds=bounds,
            attributes=self.attributes,
            coord_system=self.coord_system,
        )
        return coord

    def update(self, old_coord, new_coord=None):
        """
        Notifies the factory of the removal/replacement of a coordinate
        which might be a dependency.

        Args:

        * old_coord:
            The coordinate to be removed/replaced.
        * new_coord:
            If None, any dependency using old_coord is removed, otherwise
            any dependency using old_coord is updated to use new_coord.

        """
        new_dependencies = self.dependencies
        for name, coord in self.dependencies.items():
            if old_coord is coord:
                new_dependencies[name] = new_coord
                try:
                    self._check_dependencies(**new_dependencies)
                except ValueError as e:
                    msg = "Failed to update dependencies. " + str(e)
                    raise ValueError(msg)
                else:
                    setattr(self, name, new_coord)
                break


class OceanSFactory(AuxCoordFactory):
    """Defines an Ocean s-coordinate factory."""

    def __init__(
        self, s=None, eta=None, depth=None, a=None, b=None, depth_c=None
    ):
        """
        Creates an Ocean s-coordinate factory with the formula:

        z(n,k,j,i) = eta(n,j,i)*(1+s(k)) + depth_c*s(k) +
                     (depth(j,i)-depth_c)*C(k)

        where:
            C(k) = (1-b) * sinh(a*s(k)) / sinh(a) +
                   b * [tanh(a * (s(k) + 0.5)) / (2 * tanh(0.5*a)) - 0.5]

        """
        # Configure the metadata manager.
        self._metadata_manager = metadata_manager_factory(CoordMetadata)
        super().__init__()

        # Check that provided coordinates meet necessary conditions.
        self._check_dependencies(s, eta, depth, a, b, depth_c)
        self.units = depth.units

        self.s = s
        self.eta = eta
        self.depth = depth
        self.a = a
        self.b = b
        self.depth_c = depth_c

        self.standard_name = "sea_surface_height_above_reference_ellipsoid"
        self.attributes = {"positive": "up"}

    @staticmethod
    def _check_dependencies(s, eta, depth, a, b, depth_c):
        # Check for sufficient factory coordinates.
        if (
            eta is None
            or s is None
            or depth is None
            or a is None
            or b is None
            or depth_c is None
        ):
            msg = (
                "Unable to construct Ocean s-coordinate "
                "factory due to insufficient source coordinates."
            )
            raise ValueError(msg)

        # Check bounds and shape.
        if s is not None and s.nbounds not in (0, 2):
            msg = (
                "Invalid s coordinate {!r}: must have either "
                "0 or 2 bounds.".format(s.name())
            )
            raise ValueError(msg)

        coords = ((eta, "eta"), (depth, "depth"))
        for coord, term in coords:
            if coord is not None and coord.nbounds:
                msg = (
                    "The {} coordinate {!r} has bounds. "
                    "These are being disregarded.".format(term, coord.name())
                )
                warnings.warn(msg, UserWarning, stacklevel=2)

        coords = ((a, "a"), (b, "b"), (depth_c, "depth_c"))
        for coord, term in coords:
            if coord is not None and coord.shape != (1,):
                msg = (
                    "Expected scalar {} coordinate {!r}: "
                    "got shape {!r}.".format(term, coord.name(), coord.shape)
                )
                raise ValueError(msg)

        # Check units.
        if s is not None and s.units.is_unknown():
            # Be graceful, and promote unknown to dimensionless units.
            s.units = cf_units.Unit("1")

        if s is not None and not s.units.is_dimensionless():
            msg = (
                "Invalid units: s coordinate {!r} "
                "must be dimensionless.".format(s.name())
            )
            raise ValueError(msg)

        coords = ((eta, "eta"), (depth, "depth"), (depth_c, "depth_c"))
        for coord, term in coords:
            if coord is not None and coord.units != depth.units:
                msg = (
                    "Incompatible units: {} coordinate {!r} and depth "
                    "coordinate {!r} must have "
                    "the same units.".format(term, coord.name(), depth.name())
                )
                raise ValueError(msg)

    @property
    def dependencies(self):
        """
        Returns a dictionary mapping from constructor argument names to
        the corresponding coordinates.

        """
        return dict(
            s=self.s,
            eta=self.eta,
            depth=self.depth,
            a=self.a,
            b=self.b,
            depth_c=self.depth_c,
        )

    def _derive(self, s, eta, depth, a, b, depth_c):
        c = (1 - b) * da.sinh(a * s) / da.sinh(a) + b * (
            da.tanh(a * (s + 0.5)) / (2 * da.tanh(0.5 * a)) - 0.5
        )
        return eta * (1 + s) + depth_c * s + (depth - depth_c) * c

    def make_coord(self, coord_dims_func):
        """
        Returns a new :class:`iris.coords.AuxCoord` as defined by this factory.

        Args:

        * coord_dims_func:
            A callable which can return the list of dimensions relevant
            to a given coordinate. See :meth:`iris.cube.Cube.coord_dims()`.

        """
        # Determine the relevant dimensions.
        derived_dims = self.derived_dims(coord_dims_func)
        dependency_dims = self._dependency_dims(coord_dims_func)

        # Build the points array.
        nd_points_by_key = self._remap(dependency_dims, derived_dims)
        points = self._derive(
            nd_points_by_key["s"],
            nd_points_by_key["eta"],
            nd_points_by_key["depth"],
            nd_points_by_key["a"],
            nd_points_by_key["b"],
            nd_points_by_key["depth_c"],
        )

        bounds = None
        if self.s.nbounds:
            # Build the bounds array.
            nd_values_by_key = self._remap_with_bounds(
                dependency_dims, derived_dims
            )
            valid_shapes = [(), (1,), (2,)]
            key = "s"
            if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                name = self.dependencies[key].name()
                msg = "Invalid bounds for {} " "coordinate {!r}.".format(
                    key, name
                )
                raise ValueError(msg)
            valid_shapes.pop()
            for key in ("eta", "depth", "a", "b", "depth_c"):
                if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                    name = self.dependencies[key].name()
                    msg = (
                        "The {} coordinate {!r} has bounds. "
                        "These are being disregarded.".format(key, name)
                    )
                    warnings.warn(msg, UserWarning, stacklevel=2)
                    # Swap bounds with points.
                    bds_shape = list(nd_points_by_key[key].shape) + [1]
                    bounds = nd_points_by_key[key].reshape(bds_shape)
                    nd_values_by_key[key] = bounds

            bounds = self._derive(
                nd_values_by_key["s"],
                nd_values_by_key["eta"],
                nd_values_by_key["depth"],
                nd_values_by_key["a"],
                nd_values_by_key["b"],
                nd_values_by_key["depth_c"],
            )

        coord = iris.coords.AuxCoord(
            points,
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            units=self.units,
            bounds=bounds,
            attributes=self.attributes,
            coord_system=self.coord_system,
        )
        return coord

    def update(self, old_coord, new_coord=None):
        """
        Notifies the factory of the removal/replacement of a coordinate
        which might be a dependency.

        Args:

        * old_coord:
            The coordinate to be removed/replaced.
        * new_coord:
            If None, any dependency using old_coord is removed, otherwise
            any dependency using old_coord is updated to use new_coord.

        """
        new_dependencies = self.dependencies
        for name, coord in self.dependencies.items():
            if old_coord is coord:
                new_dependencies[name] = new_coord
                try:
                    self._check_dependencies(**new_dependencies)
                except ValueError as e:
                    msg = "Failed to update dependencies. " + str(e)
                    raise ValueError(msg)
                else:
                    setattr(self, name, new_coord)
                break


class OceanSg2Factory(AuxCoordFactory):
    """Defines an Ocean s-coordinate, generic form 2 factory."""

    def __init__(self, s=None, c=None, eta=None, depth=None, depth_c=None):
        """
        Creates an Ocean s-coordinate, generic form 2 factory with the formula:

        z(n,k,j,i) = eta(n,j,i) + (eta(n,j,i) + depth(j,i)) * S(k,j,i)

        where:
            S(k,j,i) = (depth_c * s(k) + depth(j,i) * C(k)) /
                       (depth_c + depth(j,i))

        """
        # Configure the metadata manager.
        self._metadata_manager = metadata_manager_factory(CoordMetadata)
        super().__init__()

        # Check that provided coordinates meet necessary conditions.
        self._check_dependencies(s, c, eta, depth, depth_c)
        self.units = depth.units

        self.s = s
        self.c = c
        self.eta = eta
        self.depth = depth
        self.depth_c = depth_c

        self.standard_name = "sea_surface_height_above_reference_ellipsoid"
        self.attributes = {"positive": "up"}

    @staticmethod
    def _check_dependencies(s, c, eta, depth, depth_c):
        # Check for sufficient factory coordinates.
        if (
            eta is None
            or s is None
            or c is None
            or depth is None
            or depth_c is None
        ):
            msg = (
                "Unable to construct Ocean s-coordinate, generic form 2 "
                "factory due to insufficient source coordinates."
            )
            raise ValueError(msg)

        # Check bounds and shape.
        coords = ((s, "s"), (c, "c"))
        for coord, term in coords:
            if coord is not None and coord.nbounds not in (0, 2):
                msg = (
                    "Invalid {} coordinate {!r}: must have either "
                    "0 or 2 bounds.".format(term, coord.name())
                )
                raise ValueError(msg)

        if s and s.nbounds != c.nbounds:
            msg = (
                "The s coordinate {!r} and c coordinate {!r} "
                "must be equally bounded.".format(s.name(), c.name())
            )
            raise ValueError(msg)

        coords = ((eta, "eta"), (depth, "depth"))
        for coord, term in coords:
            if coord is not None and coord.nbounds:
                msg = (
                    "The {} coordinate {!r} has bounds. "
                    "These are being disregarded.".format(term, coord.name())
                )
                warnings.warn(msg, UserWarning, stacklevel=2)

        if depth_c is not None and depth_c.shape != (1,):
            msg = (
                "Expected scalar depth_c coordinate {!r}: "
                "got shape {!r}.".format(depth_c.name(), depth_c.shape)
            )
            raise ValueError(msg)

        # Check units.
        coords = ((s, "s"), (c, "c"))
        for coord, term in coords:
            if coord is not None and coord.units.is_unknown():
                # Be graceful, and promote unknown to dimensionless units.
                coord.units = cf_units.Unit("1")

            if coord is not None and not coord.units.is_dimensionless():
                msg = (
                    "Invalid units: {} coordinate {!r} "
                    "must be dimensionless.".format(term, coord.name())
                )
                raise ValueError(msg)

        coords = ((eta, "eta"), (depth, "depth"), (depth_c, "depth_c"))
        for coord, term in coords:
            if coord is not None and coord.units != depth.units:
                msg = (
                    "Incompatible units: {} coordinate {!r} and depth "
                    "coordinate {!r} must have "
                    "the same units.".format(term, coord.name(), depth.name())
                )
                raise ValueError(msg)

    @property
    def dependencies(self):
        """
        Returns a dictionary mapping from constructor argument names to
        the corresponding coordinates.

        """
        return dict(
            s=self.s,
            c=self.c,
            eta=self.eta,
            depth=self.depth,
            depth_c=self.depth_c,
        )

    def _derive(self, s, c, eta, depth, depth_c):
        S = (depth_c * s + depth * c) / (depth_c + depth)
        return eta + (eta + depth) * S

    def make_coord(self, coord_dims_func):
        """
        Returns a new :class:`iris.coords.AuxCoord` as defined by this factory.

        Args:

        * coord_dims_func:
            A callable which can return the list of dimensions relevant
            to a given coordinate. See :meth:`iris.cube.Cube.coord_dims()`.

        """
        # Determine the relevant dimensions.
        derived_dims = self.derived_dims(coord_dims_func)
        dependency_dims = self._dependency_dims(coord_dims_func)

        # Build the points array.
        nd_points_by_key = self._remap(dependency_dims, derived_dims)
        points = self._derive(
            nd_points_by_key["s"],
            nd_points_by_key["c"],
            nd_points_by_key["eta"],
            nd_points_by_key["depth"],
            nd_points_by_key["depth_c"],
        )

        bounds = None
        if self.s.nbounds or (self.c and self.c.nbounds):
            # Build the bounds array.
            nd_values_by_key = self._remap_with_bounds(
                dependency_dims, derived_dims
            )
            valid_shapes = [(), (1,), (2,)]
            key = "s"
            if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                name = self.dependencies[key].name()
                msg = "Invalid bounds for {} " "coordinate {!r}.".format(
                    key, name
                )
                raise ValueError(msg)
            valid_shapes.pop()
            for key in ("eta", "depth", "depth_c"):
                if nd_values_by_key[key].shape[-1:] not in valid_shapes:
                    name = self.dependencies[key].name()
                    msg = (
                        "The {} coordinate {!r} has bounds. "
                        "These are being disregarded.".format(key, name)
                    )
                    warnings.warn(msg, UserWarning, stacklevel=2)
                    # Swap bounds with points.
                    bds_shape = list(nd_points_by_key[key].shape) + [1]
                    bounds = nd_points_by_key[key].reshape(bds_shape)
                    nd_values_by_key[key] = bounds

            bounds = self._derive(
                nd_values_by_key["s"],
                nd_values_by_key["c"],
                nd_values_by_key["eta"],
                nd_values_by_key["depth"],
                nd_values_by_key["depth_c"],
            )

        coord = iris.coords.AuxCoord(
            points,
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            units=self.units,
            bounds=bounds,
            attributes=self.attributes,
            coord_system=self.coord_system,
        )
        return coord

    def update(self, old_coord, new_coord=None):
        """
        Notifies the factory of the removal/replacement of a coordinate
        which might be a dependency.

        Args:

        * old_coord:
            The coordinate to be removed/replaced.
        * new_coord:
            If None, any dependency using old_coord is removed, otherwise
            any dependency using old_coord is updated to use new_coord.

        """
        new_dependencies = self.dependencies
        for name, coord in self.dependencies.items():
            if old_coord is coord:
                new_dependencies[name] = new_coord
                try:
                    self._check_dependencies(**new_dependencies)
                except ValueError as e:
                    msg = "Failed to update dependencies. " + str(e)
                    raise ValueError(msg)
                else:
                    setattr(self, name, new_coord)
                break
